"""TcEx Framework Module"""
# third-party
from pydantic import BaseModel, Field


class PlaybookVariableModel(BaseModel):
    """Playbook Variable Model

    Model for the individual parts of a playbook variable -> #App:1234:output!String.
    """

    app_type: str = Field(
        ...,
        description='The application type (e.g., App|Trigger).',
    )
    job_id: str = Field(
        ...,
        description='The job id.',
    )
    key: str = Field(
        ...,
        description='The variable key (e.g., app.api_token).',
    )
    type: str = Field(
        ...,
        description='The specific variable type (e.g., String, StringArray etc).',
    )



"""TcEx Framework Module"""
# standard library
import logging

# third-party
from requests import Response, Session  # TYPE-CHECKING

# first-party
from tcex.exit.error_code import handle_error
from tcex.logger.trace_logger import TraceLogger

# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class DataStore:
    """TcEx DataStore Class.

    **domain**

    This can be either “system”, “organization”, or “local”. Any string not defined as these
    3 will return a “Resource Not Found” error response.

    **dataType**

    This is a free form index type name as allowed by ElasticSearch. The ThreatConnect API will
    use this resource verbatim for ElasticSearch calls.

    **searchCommand**

    The URI string after the typeName is assumed to be ElasticSearch relevant information and is
    passed directly to ElasticSearch.

    Args:
        session_tc: A requests.Session instance with auth configured for the ThreatConnect API.
        domain: A value of “system”, “organization”, or “local”.
        data_type: A free form type name for the data.
        mapping: Elasticsearch mapping data.
    """

    def __init__(
        self, session_tc: Session, domain: str, data_type: str, mapping: dict | None = None
    ):
        """Initialize instance properties."""
        self.domain = domain
        self.data_type = data_type
        self.mapping = mapping or {'dynamic': False}
        self.session_tc = session_tc

        # properties
        self.log = _logger

        # setup
        self._create_index()  # create the initial index.
        self._update_mappings()  # update mappings

    def _create_index(self):
        """Create index if it doesn't exist."""
        if not self.index_exists:
            rid = 'temp-index-create'
            index_settings = {}
            headers = {'Content-Type': 'application/json', 'DB-Method': 'POST'}
            url = f'/v2/exchange/db/{self.domain}/{self.data_type}/{rid}'
            r: 'Response' = self.session_tc.post(url, json=index_settings, headers=headers)

            if not r.ok:
                error: str = r.text or r.reason
                handle_error(code=800, message_values=[r.status_code, error])
            self.log.debug(f'creating index. status_code: {r.status_code}, response: "{r.text}".')
            # delete temporary record
            self.delete(rid, False)

    def _update_mappings(self):
        """Update the mapping for the current index."""
        headers = {'Content-Type': 'application/json', 'DB-Method': 'PUT'}
        url = f'/v2/exchange/db/{self.domain}/{self.data_type}/_mappings'
        r: 'Response' = self.session_tc.post(url, json=self.mapping, headers=headers)
        self.log.debug(f'update mapping. status_code: {r.status_code}, response: "{r.text}".')

    @property
    def index_exists(self) -> bool:
        """Check to see if index exists."""
        headers = {'Content-Type': 'application/json', 'DB-Method': 'GET'}
        url = f'/v2/exchange/db/{self.domain}/{self.data_type}/_search'
        r: 'Response' = self.session_tc.post(url, headers=headers)
        if not r.ok:
            self.log.warning(f'The provided index was not found ({r.text}).')
            return False
        return True

    def add(self, rid: str, data: dict, raise_on_error: bool = True) -> dict | None:
        """Write data to the DataStore. Alias for post() method.

        Args:
            rid: The record identifier.
            data: The record data.
            raise_on_error: If True and not r.ok this method will raise a RunTimeError.

        Returns:
            dict : The response data.
        """
        return self.post(rid, data, raise_on_error)

    def delete(self, rid: str, raise_on_error: bool = True) -> dict | None:
        """Delete a record from the index using provide Id.

        **Example Response**

        .. code-block:: json

            {
              "_index": "$local.usr5_pytest",
              "_type": "pytest",
              "_id": "three",
              "_version": 4,
              "result": "deleted",
              "_shards": {
                "total": 2,
                "successful": 1,
                "failed": 0
              },
              "_seq_no": 45,
              "_primary_term": 1
            }

        Args:
            rid: The record identifier.
            raise_on_error: If True and not r.ok this method will raise a RunTimeError.

        Returns:
            dict : The response data.
        """
        response_data: dict | None = None
        headers = {'Content-Type': 'application/json', 'DB-Method': 'DELETE'}
        url = f'/v2/exchange/db/{self.domain}/{self.data_type}/{rid}'
        r: 'Response' = self.session_tc.post(url, headers=headers)
        self.log.debug(f'datastore delete status code: {r.status_code}')
        if r.ok and 'application/json' in r.headers.get('content-type', ''):
            response_data = r.json()
        else:
            error: str = r.text or r.reason
            handle_error(
                code=805,
                message_values=['delete', r.status_code, error],
                raise_error=raise_on_error,
            )
        return response_data

    def get(
        self,
        rid: str | None = None,
        data: dict | None = None,
        raise_on_error: bool = True,
    ) -> dict | None:
        """Get data from the DataStore.

        **Example Response**

        .. code-block:: json

            {
              "_index": "$local.usr5_pytest",
              "_type": "pytest",
              "_id": "two",
              "_version": 1,
              "found": true,
              "_source": {
                "two": 2
              }
            }

        Args:
            rid: The record identifier.
            data: The search query.
            raise_on_error: If True and not r.ok this method will raise a RunTimeError.

        Returns:
            dict : Python request response.
        """
        response_data: dict | None = None
        headers = {'Content-Type': 'application/json', 'DB-Method': 'GET'}
        url = f'/v2/exchange/db/{self.domain}/{self.data_type}/'
        if rid is not None:
            url = f'/v2/exchange/db/{self.domain}/{self.data_type}/{rid}'

        r: Response = self.session_tc.post(url, json=data, headers=headers)
        self.log.debug(f'datastore get status code: {r.status_code}')
        if 'application/json' in r.headers.get('content-type', ''):
            # as long as the content is JSON set the value
            try:
                response_data = r.json()
            except Exception as e:  # pragma: no cover
                # This issue should be addressed by core in a future release.
                self.log.warning(
                    'DataStore API returned a non-JSON response, even though '
                    f'content-type was application/json: {r.content} error: {e}'
                )
        if not r.ok:
            error: str = r.text or r.reason
            handle_error(
                code=805,
                message_values=['get', r.status_code, error],
                raise_error=raise_on_error,
            )
        return response_data

    def post(self, rid: str, data: dict, raise_on_error: bool = True) -> dict | None:
        """Write data to the DataStore.

        **Example Response**

        .. code-block:: json

            {
              "_index": "$local.usr5_pytest",
              "_type": "pytest",
              "_id": "one",
              "_version": 31,
              "result": "updated",
              "_shards": {
                "total": 2,
                "successful": 1,
                "failed": 0
              },
              "_seq_no": 53,
              "_primary_term": 1
            }

        Args:
            rid: The record identifier. If a value of None is passed an identifier elastic
                search will automatically generate a id.
            data: The record data.
            raise_on_error: If True and not r.ok this method will raise a RunTimeError.

        Returns:
            dict : The response data.
        """
        response_data: dict | None = None
        headers = {'Content-Type': 'application/json', 'DB-Method': 'POST'}
        url = f'/v2/exchange/db/{self.domain}/{self.data_type}/'
        if rid is not None:
            url = f'{url}{rid}'

        r: Response = self.session_tc.post(url, json=data, headers=headers)
        self.log.debug(f'datastore post status code: {r.status_code}')

        if r.ok and 'application/json' in r.headers.get('content-type', ''):
            response_data = r.json()
        else:
            error = r.text or r.reason
            handle_error(
                code=805,
                message_values=['post', r.status_code, error],
                raise_error=raise_on_error,
            )
        return response_data

    def put(self, rid: str, data: dict, raise_on_error: bool = True) -> dict | None:
        """Update the data for the provided Id.

        **Example Response**

        .. code-block:: json

            {
              "_index": "$local.usr5_pytest",
              "_type": "pytest",
              "_id": "one",
              "_version": 33,
              "result": "updated",
              "_shards": {
                "total": 2,
                "successful": 1,
                "failed": 0
              },
              "_seq_no": 55,
              "_primary_term": 1
            }

        Args:
            rid: The record identifier.
            data: A search query
            raise_on_error: If True and not r.ok this method will raise a RunTimeError.

        Returns:
            dict : The response dict.
        """
        response_data: dict | None = None
        headers = {'Content-Type': 'application/json', 'DB-Method': 'PUT'}
        url = f'/v2/exchange/db/{self.domain}/{self.data_type}/{rid}'

        r: 'Response' = self.session_tc.post(url, json=data, headers=headers)
        self.log.debug(f'datastore put status code: {r.status_code}')

        if r.ok and 'application/json' in r.headers.get('content-type', ''):
            response_data = r.json()
        else:
            error: str = r.text or r.reason
            handle_error(
                code=805,
                message_values=['put', r.status_code, error],
                raise_error=raise_on_error,
            )
        return response_data

    def update(self, rid: str, data: dict, raise_on_error: bool = True) -> dict | None:
        """Update the for the provided Id. Alias for put() method.

        Args:
            rid: The record identifier.
            data: The record data.
            raise_on_error: If True and not r.ok this method will raise a RunTimeError.

        Returns:
            dict : The response data.
        """
        return self.put(rid, data, raise_on_error)



"""TcEx Framework Module"""
# standard library
import datetime
import logging
from collections.abc import Callable
from typing import Any

# third-party
import wrapt

# first-party
from tcex.logger.trace_logger import TraceLogger

# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class Benchmark:
    """Log benchmarking times.

    This decorator will log the time of execution (benchmark_time) to the app.log file. It can be
    helpful in troubleshooting performance issues with Apps.

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        import time

        @Benchmark()
        def my_method():
            time.sleep(1)
    """

    def __init__(
        self,
        microseconds: int = 0,
        milliseconds: int = 0,
        seconds: int = 0,
    ):
        """Initialize instance properties."""
        self.microseconds = microseconds
        self.milliseconds = milliseconds
        self.seconds = seconds

    @wrapt.decorator
    def __call__(self, *wrapped_args) -> Any:
        """Implement __call__ function for decorator.

        Args:
            wrapped (callable): The wrapped function which in turns
                needs to be called by your wrapper function.
            instance (App): The object to which the wrapped
                function was bound when it was called.
            args (list): The list of positional arguments supplied
                when the decorated function was called.
            kwargs (dict): The dictionary of keyword arguments
                supplied when the decorated function was called.

        Returns:
            function: The custom decorator function.
        """
        # using wrapped args to support typing hints in PyRight
        wrapped: Callable = wrapped_args[0]
        args: list = wrapped_args[2] if len(wrapped_args) > 1 else []
        kwargs: dict = wrapped_args[3] if len(wrapped_args) > 2 else {}

        def benchmark() -> Any:
            """Iterate over data, calling the decorated function for each value.

            Args:
                app (class): The instance of the App class "self".
                *args: Additional positional arguments.
                **kwargs: Additional keyword arguments.
            """

            before = datetime.datetime.now()
            data = wrapped(*args, **kwargs)
            after = datetime.datetime.now()
            delta = after - before
            if delta > datetime.timedelta(
                microseconds=self.microseconds, milliseconds=self.milliseconds, seconds=self.seconds
            ):
                _logger.debug(f'function: "{wrapped.__name__}", benchmark_time: "{delta}"')
            return data

        return benchmark()



"""TcEx Framework Module"""
# pylint: disable=no-member,no-self-argument,wrong-import-position
# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class OwnerRoleModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='OwnerRole Model',
    validate_assignment=True,
):
    """Owner_Role Model"""

    _associated_type = PrivateAttr(False)
    _cm_type = PrivateAttr(False)
    _shared_type = PrivateAttr(False)
    _staged = PrivateAttr(False)

    available: bool = Field(
        None,
        allow_mutation=False,
        description='The **available** for the Owner_Role.',
        read_only=True,
        title='available',
    )
    comm_role: bool = Field(
        None,
        allow_mutation=False,
        description='The **comm role** for the Owner_Role.',
        read_only=True,
        title='commRole',
    )
    description_admin: str | None = Field(
        None,
        allow_mutation=False,
        description='The **description admin** for the Owner_Role.',
        read_only=True,
        title='descriptionAdmin',
    )
    description_comm: str | None = Field(
        None,
        allow_mutation=False,
        description='The **description comm** for the Owner_Role.',
        read_only=True,
        title='descriptionComm',
    )
    description_org: str | None = Field(
        None,
        allow_mutation=False,
        description='The **description org** for the Owner_Role.',
        read_only=True,
        title='descriptionOrg',
    )
    id: int | None = Field(
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    name: str | None = Field(
        None,
        allow_mutation=False,
        description='The **name** for the Owner_Role.',
        read_only=True,
        title='name',
    )
    org_role: bool = Field(
        None,
        allow_mutation=False,
        description='The **org role** for the Owner_Role.',
        read_only=True,
        title='orgRole',
    )
    version: int | None = Field(
        None,
        allow_mutation=False,
        description='The **version** for the Owner_Role.',
        read_only=True,
        title='version',
    )


class OwnerRoleDataModel(
    BaseModel,
    title='OwnerRole Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Owner_Roles Data Model"""

    data: list[OwnerRoleModel] | None = Field(
        [],
        description='The data for the OwnerRoles.',
        methods=['POST', 'PUT'],
        title='data',
    )


class OwnerRolesModel(
    BaseModel,
    title='OwnerRoles Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Owner_Roles Model"""

    _mode_support = PrivateAttr(False)

    data: list[OwnerRoleModel] | None = Field(
        [],
        description='The data for the OwnerRoles.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# add forward references
OwnerRoleDataModel.update_forward_refs()
OwnerRoleModel.update_forward_refs()
OwnerRolesModel.update_forward_refs()



"""TcEx Framework Module"""
# standard library
from collections.abc import Generator

# first-party
from tcex.input.field_type.edit_choice import EditChoice
from tcex.input.field_type.exception import InvalidInput


class Choice(EditChoice):
    """Choice Field Type"""

    @classmethod
    def __get_validators__(cls) -> Generator:
        """Run validators / modifiers on input."""
        yield from super().__get_validators__()
        yield cls.modifier_select

    @classmethod
    def modifier_select(cls, value: str, field) -> str | None:
        """Modify value if -- Select -- option is selected.

        Job Apps: If not selection None is sent.
        PB Apps: '-- Select --' has to be added to validValues by developer.
        """
        _value = value
        if _value == '-- Select --':
            if field.allow_none is False:
                raise InvalidInput(field.name, f'{_value} is not a valid choice.')
            _value = None

        return _value


def choice(value_transformations: dict[str, str] | None = None) -> type[Choice]:
    """Return configured instance of String.

    :param value_transformations: dictionary that dictates how a choice should be transformed.
    Dictionary keys should be the field's valid values as defined in the install.json. Example:

    value_transformations: {'my_choice': 'My Choice'}

    If this field were to be initialized with 'my_choice', then the final value found in the input
    model would be 'My Choice'.
    """
    namespace = {'_value_transformations': value_transformations}
    return type('CustomChoice', (Choice,), namespace)



"""TcEx Framework Module"""
# standard library
from collections.abc import Callable
from functools import cached_property as functool_cached_property
from typing import Generic, TypeVar

R = TypeVar('R')


class cached_property(functool_cached_property, Generic[R]):
    """Customized cached_property."""

    # pylint: disable=useless-super-delegation
    def __init__(self, func: Callable[..., R]):
        """Initialize instance properties."""
        super().__init__(func)

    instances = []

    def __get__(self, instance, owner=None) -> R:
        """Override method."""
        self.instances.append(instance)
        return super().__get__(instance, owner)  # type: ignore

    @staticmethod
    def _reset():
        """Reset cache"""
        for i in cached_property.instances:
            for key, value in i.__dict__.items():
                if isinstance(value, cached_property):
                    del i.__dict__[key]



"""TcEx Framework Module"""
# standard library
from enum import Enum


class ApiEndpoints(Enum):
    """Available API Endpoints"""

    ARTIFACT_TYPES = '/v3/artifactTypes'
    ATTRIBUTE_TYPES = '/v3/attributeTypes'
    ARTIFACTS = '/v3/artifacts'
    CASES = '/v3/cases'
    CASE_ATTRIBUTES = '/v3/caseAttributes'
    GROUP_ATTRIBUTES = '/v3/groupAttributes'
    GROUPS = '/v3/groups'
    INDICATOR_ATTRIBUTES = '/v3/indicatorAttributes'
    INDICATORS = '/v3/indicators'
    NOTES = '/v3/notes'
    OWNERS = '/v3/security/owners'
    OWNER_ROLES = '/v3/security/ownerRoles'
    SECURITY_LABELS = '/v3/securityLabels'
    SYSTEM_ROLES = '/v3/security/systemRoles'
    TAGS = '/v3/tags'
    TASKS = '/v3/tasks'
    USER_GROUPS = '/v3/security/userGroups'
    USERS = '/v3/security/users'
    VICTIMS = '/v3/victims'
    VICTIM_ASSETS = '/v3/victimAssets'
    VICTIM_ATTRIBUTES = '/v3/victimAttributes'
    WORKFLOW_EVENTS = '/v3/workflowEvents'
    WORKFLOW_TEMPLATES = '/v3/workflowTemplates'



"""TcEx Framework Module"""
# first-party
from tcex.requests_external.external_session import ExternalSession
from tcex.requests_external.requests_external import RequestsExternal

__all__ = ['ExternalSession', 'RequestsExternal']



"""TcEx Framework Module"""
# standard library
import json
from collections.abc import Generator, Iterator
from datetime import datetime
from typing import TYPE_CHECKING, Self

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.artifacts.artifact_model import ArtifactModel
from tcex.api.tc.v3.cases.case_model import CaseModel
from tcex.api.tc.v3.file_actions.file_action_model import FileActionModel
from tcex.api.tc.v3.file_occurrences.file_occurrence_model import FileOccurrenceModel
from tcex.api.tc.v3.groups.group_model import GroupModel
from tcex.api.tc.v3.indicator_attributes.indicator_attribute_model import IndicatorAttributeModel
from tcex.api.tc.v3.indicators.indicator_filter import IndicatorFilter
from tcex.api.tc.v3.indicators.indicator_model import IndicatorModel, IndicatorsModel
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.security_labels.security_label_model import SecurityLabelModel
from tcex.api.tc.v3.tags.tag_model import TagModel

if TYPE_CHECKING:  # pragma: no cover
    # first-party
    from tcex.api.tc.v3.artifacts.artifact import Artifact  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.cases.case import Case  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.groups.group import Group  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.indicator_attributes.indicator_attribute import (  # CIRCULAR-IMPORT
        IndicatorAttribute,
    )
    from tcex.api.tc.v3.security_labels.security_label import SecurityLabel  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.tags.tag import Tag  # CIRCULAR-IMPORT


class Indicator(ObjectABC):
    """Indicators Object.

    Args:
        active (bool, kwargs): Is the indicator active?
        active_locked (bool, kwargs): Lock the indicator active value?
        address (str, kwargs): The email address associated with this indicator (EmailAddress
            specific summary field).
        associated_artifacts (Artifacts, kwargs): A list of Artifacts associated with this
            Indicator.
        associated_cases (Cases, kwargs): A list of Cases associated with this Indicator.
        associated_groups (Groups, kwargs): A list of groups that this indicator is associated with.
        associated_indicators (Indicators, kwargs): A list of indicators associated with this
            indicator.
        attributes (IndicatorAttributes, kwargs): A list of Attributes corresponding to the
            Indicator.
        confidence (int, kwargs): The indicator threat confidence.
        dns_active (bool, kwargs): Is dns active for the indicator?
        file_actions (FileActions, kwargs): The type of file action associated with this indicator.
        file_occurrences (FileOccurrences, kwargs): A list of file occurrences associated with this
            indicator.
        host_name (str, kwargs): The host name of the indicator (Host specific summary field).
        ip (str, kwargs): The ip address associated with this indicator (Address specific summary
            field).
        md5 (str, kwargs): The md5 associated with this indicator (File specific summary field).
        mode (str, kwargs): The operation to perform on the file hashes (delete | merge).
        owner_id (int, kwargs): The id of the Organization, Community, or Source that the item
            belongs to.
        owner_name (str, kwargs): The name of the Organization, Community, or Source that the item
            belongs to.
        private_flag (bool, kwargs): Is this indicator private?
        rating (int, kwargs): The indicator threat rating.
        security_labels (SecurityLabels, kwargs): A list of Security Labels corresponding to the
            Intel item (NOTE: Setting this parameter will replace any existing tag(s) with
            the one(s) specified).
        sha1 (str, kwargs): The sha1 associated with this indicator (File specific summary field).
        sha256 (str, kwargs): The sha256 associated with this indicator (File specific summary
            field).
        size (int, kwargs): The size of the file.
        tags (Tags, kwargs): A list of Tags corresponding to the item (NOTE: Setting this parameter
            will replace any existing tag(s) with the one(s) specified).
        text (str, kwargs): The url text value of the indicator (Url specific summary field).
        type (str, kwargs): The **type** for the Indicator.
        value1 (str, kwargs): Custom Indicator summary field value1.
        value2 (str, kwargs): Custom Indicator summary field value2.
        value3 (str, kwargs): Custom Indicator summary field value3.
        whois_active (bool, kwargs): Is whois active for the indicator?
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model: IndicatorModel = IndicatorModel(**kwargs)
        self._nested_field_name = 'associatedIndicators'
        self._nested_filter = 'has_indicator'
        self.type_ = 'Indicator'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.INDICATORS.value

    @property
    def model(self) -> IndicatorModel:
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: dict | IndicatorModel):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            raise RuntimeError(f'Invalid data type: {type(data)} provided.')

    @property
    def as_entity(self) -> dict:
        """Return the entity representation of the object."""
        type_ = self.model.type

        return {'type': type_, 'id': self.model.id, 'value': self.model.summary}

    def remove(self, params: dict | None = None):
        """Remove a nested object."""
        method = 'PUT'
        unique_id = self._calculate_unique_id()

        # validate an id is available
        self._validate_id(unique_id.get('value'), '')

        body = json.dumps(
            {
                self._nested_field_name: {
                    'data': [{unique_id.get('filter'): unique_id.get('value')}],
                    'mode': 'delete',
                }
            }
        )

        # get the unique id value for id, xid, summary, etc ...
        parent_api_endpoint = self._parent_data.get('api_endpoint')
        parent_unique_id = self._parent_data.get('unique_id')
        url = f'{parent_api_endpoint}/{parent_unique_id}'

        # validate parent an id is available
        self._validate_id(parent_unique_id, url)

        self._request(
            method=method,
            url=url,
            body=body,
            headers={'content-type': 'application/json'},
            params=params,
        )

        return self.request

    @property
    def associated_artifacts(self) -> Generator['Artifact', None, None]:
        """Yield Artifact from Artifacts."""
        # first-party
        from tcex.api.tc.v3.artifacts.artifact import Artifacts

        yield from self._iterate_over_sublist(Artifacts)  # type: ignore

    @property
    def associated_cases(self) -> Generator['Case', None, None]:
        """Yield Case from Cases."""
        # first-party
        from tcex.api.tc.v3.cases.case import Cases

        yield from self._iterate_over_sublist(Cases)  # type: ignore

    @property
    def associated_groups(self) -> Generator['Group', None, None]:
        """Yield Group from Groups."""
        # first-party
        from tcex.api.tc.v3.groups.group import Groups

        yield from self._iterate_over_sublist(Groups)  # type: ignore

    @property
    def associated_indicators(self) -> Generator[Self, None, None]:
        """Yield Indicator from Indicators."""
        # Ensure the current item is not returned as a association
        for indicator in self._iterate_over_sublist(Indicators):  # type: ignore
            if indicator.model.id == self.model.id:
                continue
            yield indicator  # type: ignore

    @property
    def attributes(self) -> Generator['IndicatorAttribute', None, None]:
        """Yield Attribute from Attributes."""
        # first-party
        from tcex.api.tc.v3.indicator_attributes.indicator_attribute import IndicatorAttributes

        yield from self._iterate_over_sublist(IndicatorAttributes)  # type: ignore

    @property
    def security_labels(self) -> Generator['SecurityLabel', None, None]:
        """Yield SecurityLabel from SecurityLabels."""
        # first-party
        from tcex.api.tc.v3.security_labels.security_label import SecurityLabels

        yield from self._iterate_over_sublist(SecurityLabels)  # type: ignore

    @property
    def tags(self) -> Generator['Tag', None, None]:
        """Yield Tag from Tags."""
        # first-party
        from tcex.api.tc.v3.tags.tag import Tags

        yield from self._iterate_over_sublist(Tags)  # type: ignore

    def stage_associated_case(self, data: dict | ObjectABC | CaseModel):
        """Stage case on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = CaseModel(**data)

        if not isinstance(data, CaseModel):
            raise RuntimeError('Invalid type passed in to stage_associated_case')
        data._staged = True
        self.model.associated_cases.data.append(data)  # type: ignore

    def stage_associated_artifact(self, data: dict | ObjectABC | ArtifactModel):
        """Stage artifact on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = ArtifactModel(**data)

        if not isinstance(data, ArtifactModel):
            raise RuntimeError('Invalid type passed in to stage_associated_artifact')
        data._staged = True
        self.model.associated_artifacts.data.append(data)  # type: ignore

    def stage_associated_group(self, data: dict | ObjectABC | GroupModel):
        """Stage group on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = GroupModel(**data)

        if not isinstance(data, GroupModel):
            raise RuntimeError('Invalid type passed in to stage_associated_group')
        data._staged = True
        self.model.associated_groups.data.append(data)  # type: ignore

    def stage_attribute(self, data: dict | ObjectABC | IndicatorAttributeModel):
        """Stage attribute on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = IndicatorAttributeModel(**data)

        if not isinstance(data, IndicatorAttributeModel):
            raise RuntimeError('Invalid type passed in to stage_attribute')
        data._staged = True
        self.model.attributes.data.append(data)  # type: ignore

    def stage_file_action(self, data: dict | ObjectABC | FileActionModel):
        """Stage file_action on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = FileActionModel(**data)

        if not isinstance(data, FileActionModel):
            raise RuntimeError('Invalid type passed in to stage_file_action')
        data._staged = True
        data.indicator._staged = True
        self.model.file_actions.data.append(data)  # type: ignore

    def stage_file_occurrence(self, data: dict | ObjectABC | FileOccurrenceModel):
        """Stage file_occurrence on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = FileOccurrenceModel(**data)

        if not isinstance(data, FileOccurrenceModel):
            raise RuntimeError('Invalid type passed in to stage_file_occurrence')
        data._staged = True
        self.model.file_occurrences.data.append(data)  # type: ignore

    def stage_security_label(self, data: dict | ObjectABC | SecurityLabelModel):
        """Stage security_label on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = SecurityLabelModel(**data)

        if not isinstance(data, SecurityLabelModel):
            raise RuntimeError('Invalid type passed in to stage_security_label')
        data._staged = True
        self.model.security_labels.data.append(data)  # type: ignore

    def stage_tag(self, data: dict | ObjectABC | TagModel):
        """Stage tag on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = TagModel(**data)

        if not isinstance(data, TagModel):
            raise RuntimeError('Invalid type passed in to stage_tag')
        data._staged = True
        self.model.tags.data.append(data)  # type: ignore


class Indicators(ObjectCollectionABC):
    """Indicators Collection.

    # Example of params input
    {
        'result_limit': 100,  # Limit the retrieved results.
        'result_start': 10,  # Starting count used for pagination.
        'fields': ['caseId', 'summary']  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = IndicatorsModel(**kwargs)
        self.type_ = 'indicators'

    def __iter__(self) -> Iterator[Indicator]:
        """Return CM objects."""
        return self.iterate(base_class=Indicator)  # type: ignore

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.INDICATORS.value

    @property
    def filter(self) -> IndicatorFilter:
        """Return the type specific filter object."""
        return IndicatorFilter(self.tql)

    def deleted(
        self,
        deleted_since: datetime | str | None,
        type_: str | None = None,
        owner: str | None = None,
    ):
        """Return deleted indicators.

        This will not use the default params set on the "Indicators"
        object and instead used the params that are passed in.
        """

        if deleted_since is not None:
            deleted_since = str(
                self.util.any_to_datetime(deleted_since).strftime('%Y-%m-%dT%H:%M:%SZ')
            )

        yield from self.iterate(
            base_class=Indicator,
            api_endpoint=f'{self._api_endpoint}/deleted',
            params={'deletedSince': deleted_since, 'owner': owner, 'type': type_},
        )



"""TcEx Framework Module"""

# first-party
from tcex.api.tc.v2.batch.batch import Batch
from tcex.api.tc.v2.batch.batch_submit import BatchSubmit
from tcex.api.tc.v2.batch.batch_writer import BatchWriter, GroupType, IndicatorType

__all__ = ['Batch', 'BatchSubmit', 'BatchWriter', 'GroupType', 'IndicatorType']



"""TcEx Framework Module"""
# third-party
from pydantic import BaseModel, Field


class AotExecutionEnabledModel(BaseModel):
    """AOT Execution Enabled Model

    Feature: aotExecutionEnabled

    Supported for the following runtimeLevel:
    * Playbook
    """

    tc_action_channel: str = Field(
        None,
        description='The AOT channel to monitor for incoming action.',
        inclusion_reason='feature (aotExecutionEnabled)',
    )
    tc_aot_enabled: bool = Field(
        False,
        description='Flag to enable AOT mode for the App.',
        inclusion_reason='feature (aotExecutionEnabled)',
    )
    tc_exit_channel: str = Field(
        None,
        description='The AOT channel to send the exit message.',
        inclusion_reason='feature (aotExecutionEnabled)',
    )
    tc_terminate_seconds: int = Field(
        None,
        description='The max number of second to wait for action message.',
        inclusion_reason='feature (aotExecutionEnabled)',
    )



"""TcEx Framework Module"""
# standard library
import os
import re
from collections.abc import Generator
from pathlib import Path
from typing import Any


# pylint: disable=duplicate-bases
class _EnvPath(type(Path()), Path):  # type: ignore
    """A stub of Path with additional attribute."""

    # store for the original value passed to EnvPath
    original_value: str | None = None


class EnvPath(Path):
    """EnvPath custom pydantic model type."""

    @classmethod
    def __modify_schema__(cls, field_schema: dict[str, Any]):
        """."""
        field_schema.update(format='file-path')

    @classmethod
    def __get_validators__(cls) -> Generator:
        """."""
        yield cls.validate

    @classmethod
    def validate(cls, value: Path | str) -> Path | str:
        """Replace any environment variables in the tcex.json file."""
        if isinstance(value, Path):
            return value

        string = str(value)
        for m in re.finditer(r'\${(env|envs|local|remote):(.*?)}', string):
            try:
                full_match = m.group(0)
                env_type = m.group(1)
                env_key = m.group(2)

                if env_type != 'env':
                    raise ValueError(f'Invalid environment type found ({env_type})')
                env_value = os.getenv(env_key)
                if env_value is not None:
                    string = string.replace(full_match, env_value)
            except IndexError:
                return string

        # convert value to Path and return original value
        p = _EnvPath(os.path.expanduser(string))
        p.original_value = value
        return p



"""TcEx Framework Module"""
# pylint: disable=no-member,no-self-argument,wrong-import-position
# standard library
from datetime import datetime

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class WorkflowEventModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='WorkflowEvent Model',
    validate_assignment=True,
):
    """Workflow_Event Model"""

    _associated_type = PrivateAttr(False)
    _cm_type = PrivateAttr(True)
    _shared_type = PrivateAttr(False)
    _staged = PrivateAttr(False)

    case_id: int | None = Field(
        None,
        description='The **case id** for the Workflow_Event.',
        methods=['POST'],
        read_only=False,
        required_alt_field='caseXid',
        title='caseId',
    )
    case_xid: str | None = Field(
        None,
        description='The **case xid** for the Workflow_Event.',
        methods=['POST'],
        read_only=False,
        required_alt_field='caseId',
        title='caseXid',
    )
    date_added: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The **date added** for the Workflow_Event.',
        read_only=True,
        title='dateAdded',
    )
    deleted: bool = Field(
        None,
        allow_mutation=False,
        description='The **deleted** for the Workflow_Event.',
        read_only=True,
        title='deleted',
    )
    deleted_reason: str | None = Field(
        None,
        description='The reason for deleting the event (required input for DELETE operation only).',
        methods=['DELETE'],
        read_only=False,
        title='deletedReason',
    )
    event_date: datetime | None = Field(
        None,
        description='The time that the Event is logged.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='eventDate',
    )
    id: int | None = Field(
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    link: str | None = Field(
        None,
        allow_mutation=False,
        description='The **link** for the Workflow_Event.',
        read_only=True,
        title='link',
    )
    link_text: str | None = Field(
        None,
        allow_mutation=False,
        description='The **link text** for the Workflow_Event.',
        read_only=True,
        title='linkText',
    )
    notes: 'NotesModel' = Field(
        None,
        description='A list of Notes corresponding to the Event.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='notes',
    )
    parent_case: 'CaseModel' = Field(
        None,
        allow_mutation=False,
        description='The **parent case** for the Workflow_Event.',
        read_only=True,
        title='parentCase',
    )
    summary: str | None = Field(
        None,
        description='The **summary** for the Workflow_Event.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='summary',
    )
    system_generated: bool = Field(
        None,
        allow_mutation=False,
        description='The **system generated** for the Workflow_Event.',
        read_only=True,
        title='systemGenerated',
    )
    user: 'UserModel' = Field(
        None,
        allow_mutation=False,
        description='The **user** for the Workflow_Event.',
        read_only=True,
        title='user',
    )

    @validator('parent_case', always=True, pre=True)
    def _validate_case(cls, v):
        if not v:
            return CaseModel()  # type: ignore
        return v

    @validator('notes', always=True, pre=True)
    def _validate_notes(cls, v):
        if not v:
            return NotesModel()  # type: ignore
        return v

    @validator('user', always=True, pre=True)
    def _validate_user(cls, v):
        if not v:
            return UserModel()  # type: ignore
        return v


class WorkflowEventDataModel(
    BaseModel,
    title='WorkflowEvent Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Workflow_Events Data Model"""

    data: list[WorkflowEventModel] | None = Field(
        [],
        description='The data for the WorkflowEvents.',
        methods=['POST', 'PUT'],
        title='data',
    )


class WorkflowEventsModel(
    BaseModel,
    title='WorkflowEvents Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Workflow_Events Model"""

    _mode_support = PrivateAttr(False)

    data: list[WorkflowEventModel] | None = Field(
        [],
        description='The data for the WorkflowEvents.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# first-party
from tcex.api.tc.v3.cases.case_model import CaseModel
from tcex.api.tc.v3.notes.note_model import NotesModel
from tcex.api.tc.v3.security.users.user_model import UserModel

# add forward references
WorkflowEventDataModel.update_forward_refs()
WorkflowEventModel.update_forward_refs()
WorkflowEventsModel.update_forward_refs()



"""TcEx Framework Module"""
# standard library
from importlib.metadata import version
from typing import TYPE_CHECKING

# third-party
from semantic_version import Version

if TYPE_CHECKING:
    from .install_json import InstallJson  # CIRCULAR-IMPORT


class InstallJsonUpdate:
    """Config object for install.json file (update)"""

    def __init__(self, ij: 'InstallJson'):  # pylint: disable=E0601
        """Initialize instance properties"""
        self.ij = ij

    def multiple(
        self,
        features: bool = True,
        # language_version: bool = True,
        migrate: bool = False,
        sequence: bool = True,
        valid_values: bool = True,
        playbook_data_types: bool = True,
        sdk_version: bool = True,
    ):
        """Update the profile with all required changes.

        Args:
            features: If True, features will be updated.
            language_version: If True, the language version will be updated.
            migrate: If True, programMain will be set to "run".
            sequence: If True, sequence numbers will be updated.
            valid_values: If True, validValues will be updated.
            playbook_data_types:  If True, pbDataTypes will be updated.
            sdk_version: If True, the sdk version will be updated.
        """
        # update features array
        if features is True:
            self.ij.model.features = self.ij.model.updated_features

        # if language_version is True:
        #     # update language version to the current version of Python
        #     self.update_language_version()

        if migrate is True:
            # update programMain to run
            self.update_program_main()

        # update sequence numbers
        if sequence is True:
            self.update_sequence_numbers()

        # update valid values
        if valid_values is True:
            self.update_valid_values()

        # update playbook data types
        if playbook_data_types is True:
            self.update_playbook_data_types()

        if sdk_version is True:
            # update language version to the current version of Python
            self.update_sdk_version()

        # write updated profile
        self.ij.write()

    # def update_language_version(self):
    #     """Update language version."""
    #     self.ij.model.language_version = Version(platform.python_version())

    def update_program_main(self):
        """Update program main."""
        self.ij.model.program_main = 'run'

    def update_sequence_numbers(self):
        """Update program sequence numbers."""
        for sequence, param in enumerate(self.ij.model.params, start=1):
            param.sequence = sequence

    def update_valid_values(self):
        """Update program main on App type."""
        for param in self.ij.model.params:
            if param.type not in ['String', 'KeyValueList']:
                continue

            store = 'TEXT'
            if param.encrypt is True:
                store = 'KEYCHAIN'

            if self.ij.model.is_organization_app or param.service_config is True:
                if f'${{USER:{store}}}' not in param.valid_values:
                    param.valid_values.append(f'${{USER:{store}}}')

                if f'${{ORGANIZATION:{store}}}' not in param.valid_values:
                    param.valid_values.append(f'${{ORGANIZATION:{store}}}')

                # remove entry that's specifically for playbooks Apps
                if f'${{{store}}}' in param.valid_values:
                    param.valid_values.remove(f'${{{store}}}')

            elif self.ij.model.is_playbook_app:
                if f'${{{store}}}' not in param.valid_values:
                    param.valid_values.append(f'${{{store}}}')

                # remove entry that's specifically for organization (job) Apps
                if f'${{USER:{store}}}' in param.valid_values:
                    param.valid_values.remove(f'${{USER:{store}}}')

                # remove entry that's specifically for organization (job) Apps
                if f'${{ORGANIZATION:{store}}}' in param.valid_values:
                    param.valid_values.remove(f'${{ORGANIZATION:{store}}}')

    def update_playbook_data_types(self):
        """Update program main on App type."""
        if not self.ij.model.is_playbook_app:
            return

        for param in self.ij.model.params:
            if param.type != 'String':
                continue
            if not param.playbook_data_type:
                param.playbook_data_type.append('String')

    def update_sdk_version(self):
        """Update sdk version."""
        try:
            # best effort to get the version of the tcex package
            self.ij.model.sdk_version = Version(version('tcex'))
        except (ImportError, ValueError):
            pass



"""TcEx Framework Module"""
# standard library
from datetime import datetime
from enum import Enum

# third-party
from arrow import Arrow

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql import Tql
from tcex.api.tc.v3.tql.tql_operator import TqlOperator
from tcex.api.tc.v3.tql.tql_type import TqlType


class ArtifactFilter(FilterABC):
    """Filter Object for Artifacts"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.ARTIFACTS.value

    def analytics_score(self, operator: Enum, analytics_score: int | list):
        """Filter Analytics Score based on **analyticsScore** keyword.

        Args:
            operator: The operator enum for the filter.
            analytics_score: The intel score of the artifact.
        """
        if isinstance(analytics_score, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('analyticsScore', operator, analytics_score, TqlType.INTEGER)

    def case_id(self, operator: Enum, case_id: int | list):
        """Filter Case ID based on **caseId** keyword.

        Args:
            operator: The operator enum for the filter.
            case_id: The ID of the case associated with this artifact.
        """
        if isinstance(case_id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('caseId', operator, case_id, TqlType.INTEGER)

    def date_added(self, operator: Enum, date_added: Arrow | datetime | int | str):
        """Filter Date Added based on **dateAdded** keyword.

        Args:
            operator: The operator enum for the filter.
            date_added: The date the artifact was added to the system.
        """
        date_added = self.util.any_to_datetime(date_added).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('dateAdded', operator, date_added, TqlType.STRING)

    @property
    def has_case(self):
        """Return **CaseFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.cases.case_filter import CaseFilter

        cases = CaseFilter(Tql())
        self._tql.add_filter('hasCase', TqlOperator.EQ, cases, TqlType.SUB_QUERY)
        return cases

    @property
    def has_group(self):
        """Return **GroupFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.groups.group_filter import GroupFilter

        groups = GroupFilter(Tql())
        self._tql.add_filter('hasGroup', TqlOperator.EQ, groups, TqlType.SUB_QUERY)
        return groups

    @property
    def has_indicator(self):
        """Return **IndicatorFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.indicators.indicator_filter import IndicatorFilter

        indicators = IndicatorFilter(Tql())
        self._tql.add_filter('hasIndicator', TqlOperator.EQ, indicators, TqlType.SUB_QUERY)
        return indicators

    @property
    def has_note(self):
        """Return **NoteFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.notes.note_filter import NoteFilter

        notes = NoteFilter(Tql())
        self._tql.add_filter('hasNote', TqlOperator.EQ, notes, TqlType.SUB_QUERY)
        return notes

    @property
    def has_task(self):
        """Return **TaskFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.tasks.task_filter import TaskFilter

        tasks = TaskFilter(Tql())
        self._tql.add_filter('hasTask', TqlOperator.EQ, tasks, TqlType.SUB_QUERY)
        return tasks

    def id(self, operator: Enum, id: int | list):  # pylint: disable=redefined-builtin
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the artifact.
        """
        if isinstance(id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def indicator_active(self, operator: Enum, indicator_active: bool):
        """Filter Active Status based on **indicatorActive** keyword.

        Args:
            operator: The operator enum for the filter.
            indicator_active: A flag indicating whether or not the artifact is active.
        """
        self._tql.add_filter('indicatorActive', operator, indicator_active, TqlType.BOOLEAN)

    def note_id(self, operator: Enum, note_id: int | list):
        """Filter Note ID based on **noteId** keyword.

        Args:
            operator: The operator enum for the filter.
            note_id: The ID of the note associated with this artifact.
        """
        if isinstance(note_id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('noteId', operator, note_id, TqlType.INTEGER)

    def source(self, operator: Enum, source: list | str):
        """Filter Source based on **source** keyword.

        Args:
            operator: The operator enum for the filter.
            source: The source of the artifact.
        """
        if isinstance(source, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('source', operator, source, TqlType.STRING)

    def summary(self, operator: Enum, summary: list | str):
        """Filter Summary based on **summary** keyword.

        Args:
            operator: The operator enum for the filter.
            summary: The summary of the artifact.
        """
        if isinstance(summary, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('summary', operator, summary, TqlType.STRING)

    def task_id(self, operator: Enum, task_id: int | list):
        """Filter Task ID based on **taskId** keyword.

        Args:
            operator: The operator enum for the filter.
            task_id: The ID of the task associated with this artifact.
        """
        if isinstance(task_id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('taskId', operator, task_id, TqlType.INTEGER)

    def type(self, operator: Enum, type: list | str):  # pylint: disable=redefined-builtin
        """Filter typeName based on **type** keyword.

        Args:
            operator: The operator enum for the filter.
            type: The type name of the artifact.
        """
        if isinstance(type, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('type', operator, type, TqlType.STRING)

    def type_name(self, operator: Enum, type_name: list | str):
        """Filter typeName based on **typeName** keyword.

        Args:
            operator: The operator enum for the filter.
            type_name: The type name of the artifact.
        """
        if isinstance(type_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('typeName', operator, type_name, TqlType.STRING)



"""TcEx Framework Module"""
# standard library
import json
from collections.abc import Callable

# first-party
from tcex.util import Util


class Attribute:
    """ThreatConnect Batch Attribute Object"""

    __slots__ = ['_attribute_data', '_valid', 'util']

    def __init__(
        self,
        attr_type: str,
        attr_value: str,
        displayed: bool = False,
        source: str | None = None,
        formatter: Callable[[str], str] | None = None,
    ):
        """Initialize instance properties.

        Args:
            attr_type: The ThreatConnect defined attribute type.
            attr_value: The value for this attribute.
            displayed: If True the supported attribute will be marked for display.
            source: The source value for this attribute.
            formatter: A callable that take a single attribute
                value and return a single formatted value.
        """
        self._attribute_data: dict[str, bool | str] = {'type': attr_type}
        if displayed:
            self._attribute_data['displayed'] = displayed

        # format the value
        if formatter is not None:
            attr_value = formatter(attr_value)
        self._attribute_data['value'] = attr_value

        # add source if provided
        if source is not None:
            self._attribute_data['source'] = source

        # is attr_value not null or ''
        self._valid = True

        # check for None and '' value only.
        if attr_value in [None, '']:
            self._valid = False

        # properties
        self.util = Util()

    @property
    def data(self) -> dict:
        """Return Attribute data."""
        return self._attribute_data

    @property
    def displayed(self) -> bool:
        """Return Attribute displayed."""
        return self.util.to_bool(self._attribute_data.get('displayed') or False)

    @displayed.setter
    def displayed(self, displayed: bool):
        """Set Attribute displayed."""
        self._attribute_data['displayed'] = displayed

    @property
    def source(self) -> str | None:
        """Return Attribute source."""
        return str(self._attribute_data.get('source'))

    @source.setter
    def source(self, source: str):
        """Set Attribute source."""
        self._attribute_data['source'] = source

    @property
    def type(self) -> str:
        """Return attribute value."""
        return str(self._attribute_data['type'])

    @property
    def valid(self) -> bool:
        """Return valid value."""
        return self._valid

    @property
    def value(self) -> bool | str | None:
        """Return attribute value."""
        return self._attribute_data['value']

    def __str__(self) -> str:
        """Return string representation of object."""
        return json.dumps(self.data, indent=4)



"""TcEx Framework Module"""
# standard library
import os
from enum import Enum
from pathlib import Path

# third-party
import typer

# first-party
from tcex.api.tc.v3._gen._gen_args_abc import GenerateArgsABC
from tcex.api.tc.v3._gen._gen_filter_abc import GenerateFilterABC
from tcex.api.tc.v3._gen._gen_model_abc import GenerateModelABC
from tcex.api.tc.v3._gen._gen_object_abc import GenerateObjectABC
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.util import Util
from tcex.util.code_operation import CodeOperation
from tcex.util.render.render import Render
from tcex.util.string_operation import SnakeString

# initialize Util module
util = Util()


def log_server():
    """Log server."""
    _api_server = os.getenv('TC_API_PATH')
    Render.panel.info(f'Using server {_api_server}')


class GenerateArgs(GenerateArgsABC):
    """Generate Models for TC API Types"""


class GenerateFilter(GenerateFilterABC):
    """Generate Models for TC API Types"""

    def __init__(self, type_: SnakeString):
        """Initialize instance properties."""
        super().__init__(type_)

        # properties
        _api_endpoint = getattr(ApiEndpoints, type_.upper()).value
        self.api_url = f'{self._api_server}{_api_endpoint}'


class GenerateModel(GenerateModelABC):
    """Generate Models for TC API Types"""

    def __init__(self, type_: SnakeString):
        """Initialize instance properties."""
        super().__init__(type_)

        # properties
        _api_endpoint = getattr(ApiEndpoints, type_.upper()).value
        self.api_url = f'{self._api_server}{_api_endpoint}'


class GenerateObject(GenerateObjectABC):
    """Generate Models for TC API Types"""

    def __init__(self, type_: SnakeString):
        """Initialize instance properties."""
        super().__init__(type_)

        # properties
        _api_endpoint = getattr(ApiEndpoints, type_.upper()).value
        self.api_url = f'{self._api_server}{_api_endpoint}'


def gen_args(type_: SnakeString, indent_blocks: int):
    """Generate args code."""
    # get instance of doc generator
    gen = GenerateArgs(type_)

    # run model code first so that requirements can be determined
    i1 = ' ' * (4 * indent_blocks)
    i2 = i1 + ' ' * 4
    print(gen.gen_args(i1=i1, i2=i2))


def gen_filter(type_: SnakeString):
    """Generate the filter code."""
    # get instance of filter generator
    gen = GenerateFilter(type_)

    # set the output filename using the appropriate tcex api path
    out_path = gen.tap(type_).split('.') + [type_, f'{type_.singular()}_filter.py']
    out_file = Path(os.path.join(*out_path))

    if not out_file.is_file():
        Render.panel.failure(f'\nCould not find file {out_file}.')

    # generate class methods first so requirements can be updated
    class_methods = gen.gen_class_methods()

    _code = gen.gen_doc_string()
    _code += gen.gen_requirements()
    _code += gen.gen_class()
    _code += class_methods
    _code += ''  # newline at the end of the file

    with out_file.open(mode='w') as fh:
        fh.write(CodeOperation.format_code(_code))

    Render.panel.success(f'Successfully wrote {out_file}.')

    for msg in gen.messages:
        Render.panel.warning(msg)


def gen_model(type_: SnakeString):
    """Generate model code."""
    # get instance of model generator
    gen = GenerateModel(type_)

    # set the output filename using the appropriate tcex api path
    out_path = gen.tap(type_).split('.') + [type_, f'{type_.singular()}_model.py']
    out_file = Path(os.path.join(*out_path))

    if not out_file.is_file():
        Render.panel.failure(f'\nCould not find file {out_file}.')

    # generate model fields code first so that requirements can be determined
    container_private_attrs = gen.gen_container_private_attrs()
    model_fields = gen.gen_model_fields()
    model_private_attrs = gen.gen_model_private_attrs()
    validator_methods = gen.gen_validator_methods()

    _code = gen.gen_doc_string()
    _code += gen.gen_requirements()
    # add generated model class (ArtifactModel, CaseModel, etc.)
    _code += gen.gen_model_class()
    _code += model_private_attrs
    _code += model_fields
    # add validators to model class
    _code += validator_methods
    # add generated "data" model class (ArtifactDataModel, CaseDataModel, etc.)
    _code += gen.gen_data_class()
    _code += gen.gen_data_fields()
    # add generated container model class (ArtifactsModel, CasesModel, etc.)
    _code += gen.gen_container_class()
    _code += container_private_attrs
    _code += gen.gen_container_fields()
    # add forward reference requirements
    _code += gen.gen_requirements_first_party_forward_reference()
    # add forward references
    _code += gen.gen_forward_reference()

    with out_file.open(mode='w') as fh:
        fh.write(CodeOperation.format_code(_code))

    Render.panel.success(f'Successfully wrote {out_file}.')

    for msg in gen.messages:
        Render.panel.warning(msg)


def gen_object(type_: SnakeString):
    """Generate object code."""
    # get instance of filter generator
    gen = GenerateObject(type_)

    # set the output filename using the appropriate tcex api path
    out_path = gen.tap(type_).split('.') + [type_, f'{type_.singular()}.py']
    out_file = Path(os.path.join(*out_path))

    if not out_file.is_file():
        Render.panel.failure(f'\nCould not find file {out_file}.')

    # generate class method code first so that requirements can be determined
    container_methods = gen.gen_container_methods()
    object_methods = gen.gen_object_methods()

    _code = gen.gen_doc_string()
    _code += gen.gen_requirements()
    # add generated class (Artifact, Case, etc.)
    _code += gen.gen_object_class()
    _code += object_methods
    # add generated container class (Artifacts, Cases, etc.)
    _code += gen.gen_container_class()
    _code += container_methods

    with out_file.open(mode='w') as fh:
        fh.write(CodeOperation.format_code(_code))

    Render.panel.success(f'Successfully wrote {out_file}.')

    for msg in gen.messages:
        Render.panel.warning(msg)


#
# CLI
#


class ObjectTypes(str, Enum):
    """Object Types"""

    # adversary_assets = 'adversary_assets'
    artifacts = 'artifacts'
    artifact_types = 'artifact_types'
    attribute_types = 'attribute_types'
    cases = 'cases'
    case_attributes = 'case_attributes'
    group_attributes = 'group_attributes'
    groups = 'groups'
    indicator_attributes = 'indicator_attributes'
    indicators = 'indicators'
    notes = 'notes'
    owner_roles = 'owner_roles'
    owners = 'owners'
    security_labels = 'security_labels'
    system_roles = 'system_roles'
    tags = 'tags'
    tasks = 'tasks'
    users = 'users'
    user_groups = 'user_groups'
    victims = 'victims'
    victim_assets = 'victim_assets'
    victim_attributes = 'victim_attributes'
    workflow_events = 'workflow_events'
    workflow_templates = 'workflow_templates'


class GenTypes(str, Enum):
    """Object Types"""

    all = 'all'
    filter = 'filter'
    model = 'model'
    object = 'object'


app = typer.Typer()


@app.command()
def all(  # pylint: disable=redefined-builtin
    gen_type: GenTypes = typer.Option(
        'all', '--gen_type', help='Generate filter, model, or object file.'
    ),
):
    """Generate Args."""
    log_server()
    gen_type_ = gen_type.value.lower()
    for type_ in ObjectTypes:
        type_ = util.snake_string(type_.value)

        if gen_type_ in ['all', 'filter']:
            gen_filter(type_)
        if gen_type_ in ['all', 'model']:
            gen_model(type_)
        if gen_type_ in ['all', 'object']:
            gen_object(type_)


@app.command()
def args(
    type_: ObjectTypes = typer.Option(
        ..., '--type', help='Generate Args for the provided object type.'
    ),
    indent_blocks: int = typer.Option(
        1, help='The number of four space blocks to indent the args.'
    ),
):
    """Generate Args."""
    log_server()
    gen_args(util.snake_string(type_.value), indent_blocks)


@app.command()
def code(
    type_: ObjectTypes = typer.Option(
        ..., '--type', help='Generate filter, model, and object files for the provided type.'
    ),
):
    """Generate Args."""
    log_server()
    tv = util.snake_string(type_.value)
    gen_filter(tv)
    gen_model(tv)
    gen_object(tv)


@app.command()
def filter(  # pylint: disable=redefined-builtin
    type_: ObjectTypes = typer.Option(
        ..., '--type', help='Generate filter object file for the provided type.'
    ),
):
    """Generate the filter code."""
    log_server()
    gen_filter(util.snake_string(type_.value))


@app.command()
def model(
    type_: ObjectTypes = typer.Option(
        ..., '--type', help='Generate model file for the provided type.'
    ),
):
    """Generate the model"""
    log_server()
    gen_model(util.snake_string(type_.value))


@app.command()
def object(  # pylint: disable=redefined-builtin
    type_: ObjectTypes = typer.Option(
        ..., '--type', help='Generate object file for the provided type.'
    ),
):
    """Generate the filter class"""
    log_server()
    gen_object(util.snake_string(type_.value))


if __name__ == '__main__':
    app()



"""TcEx Framework Module"""
# standard library
from typing import TYPE_CHECKING

# first-party
from tcex.api.tc.v2.threat_intelligence.tcex_ti_tc_request import TiTcRequest

if TYPE_CHECKING:
    # first-party
    from tcex.api.tc.v2.threat_intelligence.threat_intelligence import ThreatIntelligence


class Tags:
    """Unique API calls for Tags API Endpoints"""

    def __init__(self, ti: 'ThreatIntelligence'):
        """Initialize instance properties."""

        # properties
        self._api_entity = 'tag'
        self._api_sub_type = None
        self._api_type = None
        self._tc_requests = TiTcRequest(ti.session_tc)
        self._type = 'tags'
        self.ti = ti

    def many(self, filters=None, owners=None, params=None):
        """Get all the tags."""
        for tag in self.tc_requests.all_tags(filters=filters, owners=owners, params=params):
            yield self.ti.tag(name=tag.get('name'))

    @property
    def tc_requests(self):
        """Get the tc request object"""
        return self._tc_requests

    @tc_requests.setter
    def tc_requests(self, tc_requests):
        """Set the tc request object."""
        self._tc_requests = tc_requests



"""TcEx Framework Module"""
# standard library
from datetime import datetime

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class FileOccurrenceModel(
    V3ModelABC,
    title='File Occurrence Model',
    extra=Extra.allow,
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """File Occurrences Model"""

    date: datetime | None = Field(
        None,
        methods=['POST', 'PUT'],
        title='date',
    )
    file_name: str | None = Field(
        None,
        methods=['POST', 'PUT'],
        title='fileName',
    )
    path: str | None = Field(
        None,
        methods=['POST', 'PUT'],
        title='path',
    )
    id: int | None = Field(
        None,
        title='id',
    )


class FileOccurrencesModel(
    BaseModel,
    title='File Occurrences Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """File Occurrences Data Model"""

    _mode_support = PrivateAttr(True)

    count: int | None = Field(None, description='The number of occurrences.')

    data: list[FileOccurrenceModel] | None = Field(
        [],
        methods=['POST', 'PUT'],
        title='data',
    )

    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


FileOccurrenceModel.update_forward_refs()
FileOccurrencesModel.update_forward_refs()



"""TcEx Framework Module"""
# standard library
from enum import Enum

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql_type import TqlType


class OwnerRoleFilter(FilterABC):
    """Filter Object for OwnerRoles"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.OWNER_ROLES.value

    def available(self, operator: Enum, available: bool):
        """Filter Available based on **available** keyword.

        Args:
            operator: The operator enum for the filter.
            available: The availability status of the role.
        """
        self._tql.add_filter('available', operator, available, TqlType.BOOLEAN)

    def comm_role(self, operator: Enum, comm_role: bool):
        """Filter Community Role based on **commRole** keyword.

        Args:
            operator: The operator enum for the filter.
            comm_role: The scope of the role.
        """
        self._tql.add_filter('commRole', operator, comm_role, TqlType.BOOLEAN)

    def description_admin(self, operator: Enum, description_admin: list | str):
        """Filter Admin Description based on **descriptionAdmin** keyword.

        Args:
            operator: The operator enum for the filter.
            description_admin: The description of this role's admin access.
        """
        if isinstance(description_admin, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('descriptionAdmin', operator, description_admin, TqlType.STRING)

    def description_comm(self, operator: Enum, description_comm: list | str):
        """Filter Community Description based on **descriptionComm** keyword.

        Args:
            operator: The operator enum for the filter.
            description_comm: The description of this role's access within a community.
        """
        if isinstance(description_comm, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('descriptionComm', operator, description_comm, TqlType.STRING)

    def description_org(self, operator: Enum, description_org: list | str):
        """Filter Organization Description based on **descriptionOrg** keyword.

        Args:
            operator: The operator enum for the filter.
            description_org: The description of this role's access within an organization.
        """
        if isinstance(description_org, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('descriptionOrg', operator, description_org, TqlType.STRING)

    def id(self, operator: Enum, id: int | list):  # pylint: disable=redefined-builtin
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the user.
        """
        if isinstance(id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def name(self, operator: Enum, name: list | str):
        """Filter Name based on **name** keyword.

        Args:
            operator: The operator enum for the filter.
            name: The name of the role.
        """
        if isinstance(name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('name', operator, name, TqlType.STRING)

    def org_role(self, operator: Enum, org_role: bool):
        """Filter Organization Role based on **orgRole** keyword.

        Args:
            operator: The operator enum for the filter.
            org_role: The scope of this role.
        """
        self._tql.add_filter('orgRole', operator, org_role, TqlType.BOOLEAN)

    def version(self, operator: Enum, version: int | list):
        """Filter Version based on **version** keyword.

        Args:
            operator: The operator enum for the filter.
            version: The version number of the role.
        """
        if isinstance(version, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('version', operator, version, TqlType.INTEGER)



"""TcEx Framework Module"""
# standard library
from abc import ABC

# first-party
from tcex.api.tc.v3._gen._gen_abc import GenerateABC
from tcex.api.tc.v3._gen._gen_args_abc import GenerateArgsABC
from tcex.util.string_operation import SnakeString


class GenerateArgs(GenerateArgsABC):
    """Generate Models for TC API Types"""


class GenerateObjectABC(GenerateABC, ABC):
    """Generate Models for Case Management Types"""

    def __init__(self, type_: SnakeString):
        """Initialize instance properties."""
        super().__init__(type_)

        # properties
        self.requirements = {
            'standard library': [],
            'third-party': [],
            'first-party': [
                'from tcex.api.tc.v3.api_endpoints import ApiEndpoints',
                'from tcex.api.tc.v3.object_abc import ObjectABC',
                'from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC',
            ],
            'first-party-forward-reference': [],
            'type-checking': [],
        }

    def _gen_code_api_endpoint_property(self) -> str:
        """Return the method code.

        @property
        def _api_endpoint(self) -> str:
            '''Return the type specific API endpoint.'''
            return ApiEndpoints.ARTIFACTS.value
        """
        return '\n'.join(
            [
                f'''{self.i1}@property''',
                f'''{self.i1}def _api_endpoint(self) -> str:''',
                f'''{self.i2}"""Return the type specific API endpoint."""''',
                f'''{self.i2}return ApiEndpoints.{self.type_.upper()}.value''',
                '',
                '',
            ]
        )

    def _gen_code_container_init_method(self) -> str:
        """Return the method code.

        def __init__(self, **kwargs):
            '''Initialize instance properties.'''
            super().__init__(
                kwargs.pop('session', None),
                kwargs.pop('tql_filter', None),
                kwargs.pop('params', None)
            )
            self._model = ArtifactsModel(**kwargs)
        """
        # add method import requirement
        classes = [
            f'{self.type_.singular().pascal_case()}Model',
            f'{self.type_.plural().pascal_case()}Model',
        ]
        self.update_requirements(self.type_, f'{self.type_.singular()}_model', classes)

        model_name = f'{self.type_.plural().pascal_case()}Model'
        return '\n'.join(
            [
                f'''{self.i1}def __init__(self, **kwargs):''',
                f'''{self.i2}"""Initialize instance properties."""''',
                f'''{self.i2}super().__init__(''',
                (
                    f'''{self.i3}kwargs.pop('session', None), '''
                    f'''kwargs.pop('tql_filter', None), '''
                    f'''kwargs.pop('params', None)'''
                ),
                f'''{self.i2})''',
                f'''{self.i2}self._model = {model_name}(**kwargs)''',
                f'''{self.i2}self.type_ = \'{self.type_.plural()}\'''',
                '',
                '',
            ]
        )

    def _gen_code_container_iter_method(self) -> str:
        """Return the method code.

        def __iter__(self) -> Iterator[Artifact]:
            '''Return CM objects.'''
            return self.iterate(base_class=Artifact)
        """
        self.requirements['standard library'].append('from collections.abc import Iterator')
        return '\n'.join(
            [
                (
                    f'''{self.i1}def __iter__(self) -> '''
                    f'''Iterator[{self.type_.singular().pascal_case()}]:'''
                ),
                f'''{self.i2}"""Return CM objects."""''',
                (
                    f'''{self.i2}return self.iterate(base_class='''
                    f'''{self.type_.singular().pascal_case()})'''
                    '''  # type: ignore'''
                ),
                '',
                '',
            ]
        )

    # def _gen_code_container_iter_method(self) -> str:
    #     """Return the method code.

    #     def __iter__(self) -> Self:
    #         '''Return CM objects.'''
    #         return self
    #     """
    #     self.requirements['standard library'].append({'module': 'typing', 'imports': ['Self']})
    #     return '\n'.join(
    #         [
    #             f'''{self.i1}def __iter__(self) -> Self:''',
    #             f'''{self.i2}"""Return CM objects."""''',
    #             f'''{self.i2}return self''',
    #             '',
    #             '',
    #         ]
    #     )

    # def _gen_code_container_next_method(self) -> str:
    #     """Return the method code.

    #     def __next__(self) -> Artifact:
    #         '''Iterate over CM objects.'''
    #         return self.iterate(base_class=Artifact)
    #     """
    #     return '\n'.join(
    #         [
    #             f'''{self.i1}def __next__(self) -> {self.type_.singular().pascal_case()}:''',
    #             f'''{self.i2}"""Return next CM objects."""''',
    #             (
    #                 f'''{self.i2}for i in self.iterate(base_class='''
    #                 f'''{self.type_.singular().pascal_case()}):'''
    #             ),
    #             f'''{self.i3}return i''',
    #             '',
    #             f'''{self.i2}raise StopIteration''',
    #             '',
    #             '',
    #         ]
    #     )

    def _gen_code_container_filter_property(self) -> str:
        """Return the method code.

        @property
        def filter(self) -> ArtifactFilter:
            '''Return the type specific filter object.'''
            return ArtifactFilter(self._session, self.tql)
        """
        filter_class = f'{self.type_.singular().pascal_case()}Filter'
        self.update_requirements(self.type_, f'{self.type_.singular()}_filter', [filter_class])
        return '\n'.join(
            [
                f'''{self.i1}@property''',
                f'''{self.i1}def filter(self) -> {filter_class}:''',
                f'''{self.i2}"""Return the type specific filter object."""''',
                f'''{self.i2}return {filter_class}(self.tql)''',
                '',
                '',
            ]
        )

    def _gen_code_deleted_method(self) -> str:
        """Return the method code.

        @property
        def filter(self) -> ArtifactFilter:
            '''Return the type specific filter object.'''
            return ArtifactFilter(self._session, self.tql)
        """
        self.requirements['first-party'].append({'module': 'datetime', 'imports': ['datetime']})
        return '\n'.join(
            [
                f'''{self.i1}def deleted(''',
                f'''{self.i2}self,''',
                f'''{self.i2}deleted_since: datetime | str | None,''',
                f'''{self.i2}type_: str | None = None,''',
                f'''{self.i2}owner: str | None = None''',
                f'''{self.i1}):''',
                f'''{self.i2}"""Return deleted indicators.''',
                '',
                f'''{self.i2}This will not use the default params set on the "Indicators" ''',
                f'''{self.i2}object and instead used the params that are passed in.''',
                f'''{self.i2}"""''',
                '',
                f'''{self.i2}if deleted_since is not None:''',
                f'''{self.i3}deleted_since = str(''',
                (
                    f'''{self.i4}self.util.any_to_datetime(deleted_since)'''
                    '''.strftime('%Y-%m-%dT%H:%M:%SZ')'''
                ),
                f'''{self.i3})''',
                '',
                f'''{self.i2}yield from self.iterate(''',
                f'''{self.i3}base_class=Indicator,''',
                f'''{self.i3}api_endpoint=f'{{self._api_endpoint}}/deleted',''',
                (
                    f'''{self.i3}params={{'deletedSince': deleted_since, '''
                    ''''owner': owner, 'type': type_}'''
                ),
                f'''{self.i2})''',
                '',
            ]
        )

    def _gen_code_group_methods(self) -> str:
        """Return the method code.

        @property
        def filter(self) -> ArtifactFilter:
            '''Return the type specific filter object.'''
            return ArtifactFilter(self._session, self.tql)
        """
        # BCS
        # self.requirements['type-checking'].append('''from requests import Response''')
        self.requirements['first-party'].append('''from requests import Response''')
        return '\n'.join(
            [
                f'''{self.i1}def download(self, params: dict | None = None) -> bytes:''',
                f'''{self.i2}"""Return the document attachment for Document/Report Types."""''',
                f'''{self.i2}self._request(''',
                f'''{self.i3}method='GET',''',
                f'''{self.i3}url=f\'\'\'{{self.url('GET')}}/download\'\'\',''',
                f'''{self.i3}headers={{'Accept': 'application/octet-stream'}},''',
                f'''{self.i3}params=params,''',
                f'''{self.i2})''',
                f'''{self.i2}return self.request.content''',
                '',
                f'''{self.i1}def pdf(self, params: dict | None = None) -> bytes:''',
                f'''{self.i2}"""Return the document attachment for Document/Report Types."""''',
                f'''{self.i2}self._request(''',
                f'''{self.i3}method='GET',''',
                f'''{self.i3}body=None,''',
                f'''{self.i3}url=f\'\'\'{{self.url('GET')}}/pdf\'\'\',''',
                f'''{self.i3}headers={{'Accept': 'application/octet-stream'}},''',
                f'''{self.i3}params=params,''',
                f'''{self.i2})''',
                '',
                f'''{self.i2}return self.request.content''',
                '',
                (
                    f'''{self.i1}def upload(self, content: bytes | str, '''
                    '''params: dict | None = None) -> Response:'''
                ),
                f'''{self.i2}"""Return the document attachment for Document/Report Types."""''',
                f'''{self.i2}self._request(''',
                f'''{self.i3}method='POST',''',
                f'''{self.i3}url=f\'\'\'{{self.url('GET')}}/upload\'\'\',''',
                f'''{self.i3}body=content,''',
                f'''{self.i3}headers={{'content-type': 'application/octet-stream'}},''',
                f'''{self.i3}params=params,''',
                f'''{self.i2})''',
                f'''{self.i2}return self.request''',
                '',
                '',
            ]
        )

    def _gen_code_object_init_method(self) -> str:
        """Return the method code.

        def __init__(self, **kwargs):
            '''Initialize instance properties'''
            super().__init__(kwargs.pop('session', None))
            self._model = ArtifactModel(**kwargs)
        """
        # set nested type
        nested_field_name = self.util.snake_string(self.type_).camel_case().plural()
        if self.type_ in ['indicators', 'groups']:
            nested_field_name = (
                self.util.snake_string(f'associated_{self.type_}').camel_case().plural()
            )
        elif self.type_ in ['case_attributes', 'group_attributes', 'indicator_attributes']:
            nested_field_name = 'attributes'

        model_name = f'{self.type_.singular().pascal_case()}Model'
        return '\n'.join(
            [
                f'''{self.i1}def __init__(self, **kwargs):''',
                f'''{self.i2}"""Initialize instance properties."""''',
                f'''{self.i2}super().__init__(kwargs.pop('session', None))''',
                '',
                f'''{self.i2}# properties''',
                f'''{self.i2}self._model: {model_name} = {model_name}(**kwargs)''',
                f'''{self.i2}self._nested_field_name = '{nested_field_name}' ''',
                f'''{self.i2}self._nested_filter = 'has_{self.type_.singular()}' ''',
                f'''{self.i2}self.type_ = \'{self.type_.singular().space_case()}\'''',
                '',
                '',
            ]
        )

    def _gen_code_object_model_property(self) -> str:
        """Return the method code.

        Override the object ABC method so that the correct typing hint return
        type can be used and the IDE will provide full model hinting.

        @property
        def model(self) -> CaseModel:
            '''Return the model data.'''
            return self._model

        @model.setter
        def model(self, data: dict | IndicatorModel):
            '''Create model using the provided data.'''
            if isinstance(data, type(self.model)):
                # provided data is already a model, nothing required to change
                self._model = data
            elif isinstance(data, dict):
                # provided data is raw response, load the model
                self._model = type(self.model)(**data)
            else:
                raise RuntimeError(f'Invalid data type: {type(data)} provided.')
        """
        return '\n'.join(
            [
                f'''{self.i1}@property''',
                f'''{self.i1}def model(self) -> {self.type_.singular().pascal_case()}Model:''',
                f'''{self.i2}"""Return the model data."""''',
                f'''{self.i2}return self._model''',
                '',
                f'''{self.i1}@model.setter''',
                (
                    f'''{self.i1}def model(self, data: '''
                    f'''dict | {self.type_.singular().pascal_case()}Model):'''
                ),
                f'''{self.i2}"""Create model using the provided data."""''',
                f'''{self.i2}if isinstance(data, type(self.model)):''',
                f'''{self.i3}# provided data is already a model, nothing required to change''',
                f'''{self.i3}self._model = data''',
                f'''{self.i2}elif isinstance(data, dict):''',
                f'''{self.i3}# provided data is raw response, load the model''',
                f'''{self.i3}self._model = type(self.model)(**data)''',
                f'''{self.i2}else:''',
                f'''{self.i3}raise RuntimeError(f'Invalid data type: {{type(data)}} provided.')''',
                '',
                '',
            ]
        )

    def _gen_code_object_as_entity_property_method(self) -> str:
        """Return the method code.

        @property
        def as_entity(self) -> dict:
            '''Return the entity representation of the object.'''
            case - name
            notes - summary
            tag - name
            task - name
            workflow_event - summary
            workflow_template_model - name

            return {'type': 'Artifact', 'id': self.model.id, 'value': self.model.summary}
        """
        name_entities = [
            'artifact_types',
            'cases',
            'tags',
            'tasks',
            'workflow_templates',
            'groups',
            'victims',
        ]
        # check if groups or indicators and if so use the self.model.type else use self.type_
        value = 'self.model.summary'
        if self.type_.lower() in name_entities:
            value = 'self.model.name'

        as_entity_property_method = [
            f'''{self.i1}@property''',
            f'''{self.i1}def as_entity(self) -> dict:''',
            f'''{self.i2}"""Return the entity representation of the object."""''',
        ]
        if self.type_.lower() in ['groups', 'indicators']:
            as_entity_property_method.append(f'''{self.i2}type_ = self.model.type''')
        elif self.type_.lower() in ['victim_assets']:
            value = 'value'
            as_entity_property_method.extend(
                [
                    f'''{self.i2}value = []''',
                    '',
                    f'''{self.i2}if self.model.type is not None:''',
                    f'''{self.i3}if self.model.type.lower() == 'phone':''',
                    f'''{self.i4}if self.model.phone:''',
                    f'''{self.i5}value.append(self.model.phone)''',
                    f'''{self.i3}elif self.model.type.lower() == 'socialnetwork':''',
                    f'''{self.i4}if self.model.social_network:''',
                    f'''{self.i5}value.append(self.model.social_network)''',
                    f'''{self.i4}if self.model.account_name:''',
                    f'''{self.i5}value.append(self.model.account_name)''',
                    f'''{self.i3}elif self.model.type.lower() == 'networkaccount':''',
                    f'''{self.i4}if self.model.network_type:''',
                    f'''{self.i5}value.append(self.model.network_type)''',
                    f'''{self.i4}if self.model.account_name:''',
                    f'''{self.i5}value.append(self.model.account_name)''',
                    f'''{self.i3}elif self.model.type.lower() == 'emailaddress':''',
                    f'''{self.i4}if self.model.address_type:''',
                    f'''{self.i5}value.append(self.model.address_type)''',
                    f'''{self.i4}if self.model.address:''',
                    f'''{self.i5}value.append(self.model.address)''',
                    f'''{self.i3}elif self.model.type.lower() == 'website':''',
                    f'''{self.i4}if self.model.website:''',
                    f'''{self.i5}value.append(self.model.website)''',
                    '',
                    '',
                    f'''{self.i2}value = ' : '.join(value) if value else \'\'''',
                    f'''{self.i2}type_ = f'Victim Asset : {{self.model.type}}\'''',
                ]
            )
        else:
            as_entity_property_method.append(f'''{self.i2}type_ = self.type_''')
        as_entity_property_method.extend(
            [
                '',
                (
                    f'''{self.i2}return {{'type': type_, 'id': '''
                    f'''self.model.id, 'value': {value}}}'''
                ),
                '',
                '',
            ]
        )
        return '\n'.join(as_entity_property_method)

    def _gen_code_object_stage_type_method(self, type_: str, model_type: str | None = None) -> str:
        """Return the method code.

        def stage_artifact(self, **kwargs):
            '''Stage an Artifact on the object.

            ...
            '''
            self.model.artifacts.data.append(ArtifactModel(**kwargs))

            _code += self._gen_code_object_add_type_method('users', 'user_access')
        """
        type_ = self.util.snake_string(type_)
        model_type = self.util.snake_string(model_type or type_)
        model_reference = model_type

        # Unlike all of the other objects, on the victims model, it references 'assets' not the
        # model name 'VictimAssets'
        if type_.lower() == 'victim_assets' and self.type_.lower() == 'victims':
            model_reference = self.util.camel_string('assets')
        elif type_.lower() == 'users':
            model_type = self.util.camel_string('user_accesses')
            model_reference = 'user_access'

        # get model from map and update requirements
        model_import_data = self._module_import_data(type_)
        self.requirements['first-party'].append(
            f'''from {model_import_data.get('model_module')} '''
            f'''import {model_import_data.get('model_class')}'''
        )
        stage_method = [
            (
                f'''{self.i1}def stage_{model_type.singular()}(self, '''
                f'''data: dict | ObjectABC | {model_import_data.get('model_class')}'''
                f'''):'''
            ),
            f'''{self.i2}"""Stage {type_.singular()} on the object."""''',
            f'''{self.i2}if isinstance(data, ObjectABC):''',
            f'''{self.i3}data = data.model  # type: ignore''',
            f'''{self.i2}elif isinstance(data, dict):''',
            f'''{self.i3}data = {model_import_data.get('model_class')}(**data)''',
            '',
            f'''{self.i2}if not isinstance(data, {model_import_data.get('model_class')}):''',
            (
                f'''{self.i3}raise RuntimeError('Invalid type '''
                f'''passed in to stage_{model_type.singular()}')'''
            ),
            f'''{self.i2}data._staged = True''',
        ]
        if type_.lower() == 'file_actions' and self.type_.lower() == 'indicators':
            # The `indicator` field in the FileActionModel must be staged to be
            # submitted through the API
            stage_method.append(f'''{self.i2}data.indicator._staged = True''')

        stage_method.extend(
            [
                f'''{self.i2}self.model.{model_reference}.data.append(data)  # type: ignore''',
                '',
                '',
            ]
        )

        return '\n'.join(stage_method)

    def _gen_code_object_remove_method(self) -> str:
        """Return the method code."""
        self.requirements['standard library'].append('import json')
        return '\n'.join(
            [
                f'''{self.i1}def remove(self, params: dict | None = None):''',
                f'''{self.i2}"""Remove a nested object."""''',
                f'''{self.i2}method = \'PUT\'''',
                f'''{self.i2}unique_id = self._calculate_unique_id()''',
                '',
                f'''{self.i2}# validate an id is available''',
                f'''{self.i2}self._validate_id(unique_id.get('value'), '')''',
                '',
                f'''{self.i2}body = json.dumps(''',
                f'''{self.i3}{{''',
                f'''{self.i4}self._nested_field_name: {{''',
                f'''{self.i5}'data': [{{unique_id.get('filter'): unique_id.get('value')}}],''',
                f'''{self.i5}'mode': 'delete',''',
                f'''{self.i4}}}''',
                f'''{self.i3}}}''',
                f'''{self.i2})''',
                '',
                f'''{self.i2}# get the unique id value for id, xid, summary, etc ...''',
                f'''{self.i2}parent_api_endpoint = self._parent_data.get('api_endpoint')''',
                f'''{self.i2}parent_unique_id = self._parent_data.get('unique_id')''',
                f'''{self.i2}url = f\'{{parent_api_endpoint}}/{{parent_unique_id}}\'''',
                '',
                f'''{self.i2}# validate parent an id is available''',
                f'''{self.i2}self._validate_id(parent_unique_id, url)''',
                '',
                f'''{self.i2}self._request(''',
                f'''{self.i3}method=method,''',
                f'''{self.i3}url=url,''',
                f'''{self.i3}body=body,''',
                f'''{self.i3}headers={{'content-type': 'application/json'}},''',
                f'''{self.i3}params=params,''',
                f'''{self.i2})''',
                '',
                f'''{self.i2}return self.request''',
                '',
                '',
            ]
        )

    def _gen_code_object_stage_assignee(self) -> str:
        """Return the method code.

        def stage_assignee(
            self, type: str, data: dict | ObjectABC | AssigneeModel]
        ):
            '''Stage artifact on the object.'''
            if isinstance(data, ObjectABC):
                data = data.model
            elif type.lower() == 'user' and isinstance(data, dict):
                data = UserModel(**data)
            elif type.lower() == 'group' and isinstance(data, dict):
                data = UserGroupModel(**data)

            if not isinstance(data, UserModel | UserGroupModel):
                raise RuntimeError('Invalid type passed in to stage_assignee')
            data._staged = True
            self.model.assignee._staged = True
            self.model.assignee.type = type
            self.model.assignee.data = data
        """
        self.requirements['first-party'].extend(
            [
                'from tcex.api.tc.v3.security.user_groups.user_group_model import UserGroupModel',
                'from tcex.api.tc.v3.security.users.user_model import UserModel',
            ]
        )

        return '\n'.join(
            [
                f'''{self.i1}# pylint: disable=redefined-builtin''',
                (
                    f'''{self.i1}def stage_assignee(self, type: str, data: '''
                    f'''dict | ObjectABC | UserModel | UserGroupModel):'''
                ),
                f'''{self.i2}"""Stage artifact on the object."""''',
                f'''{self.i2}if isinstance(data, ObjectABC):''',
                f'''{self.i3}data = data.model  # type: ignore''',
                f'''{self.i2}elif type.lower() == 'user' and isinstance(data, dict):''',
                f'''{self.i3}data = UserModel(**data)''',
                f'''{self.i2}elif type.lower() == 'group' and isinstance(data, dict):''',
                f'''{self.i3}data = UserGroupModel(**data)''',
                '',
                f'''{self.i2}if not isinstance(data, UserModel | UserGroupModel):''',
                f'''{self.i3}raise RuntimeError('Invalid type passed in to stage_assignee')''',
                f'''{self.i2}data._staged = True''',
                f'''{self.i2}self.model.assignee._staged = True''',
                f'''{self.i2}self.model.assignee.type = type''',
                # pylance shows a warning on type here, but it in not handling inheritance properly.
                f'''{self.i2}self.model.assignee.data = data  # type: ignore''',
                '',
                '',
            ]
        )

    def _gen_code_object_type_property_method(
        self, type_: str, model_type: str | None = None
    ) -> str:
        """Return the method code.

        @property
        def artifacts(self):
            '''Yield Artifact from Artifacts'''
            from tcex.api.tc.v3.artifacts.model import Artifact, Artifacts

            yield from self._iterate_over_sublist(Artifacts)
        """
        type_ = self.util.snake_string(type_)
        model_type = self.util.snake_string(model_type or type_)

        # get model from map and update requirements
        model_import_data = self._module_import_data(type_)

        # Add Generator to imports:
        self.requirements['standard library'].append(
            {'module': 'collections.abc', 'imports': ['Generator']}
        )

        if self.type_ == type_:
            # set return type to new typing "Self" type and add import
            return_type = 'Self'
            self.requirements['standard library'].append(
                {'module': 'typing', 'imports': [return_type]}
            )
        else:
            # set typing to model class and add import
            return_type = f'''\'{model_import_data.get('object_class')}\''''
            self.requirements['type-checking'].append(
                f'''from {model_import_data.get('object_module')} '''
                f'''import {model_import_data.get('object_class')}'''
            )

        _code = [
            f'''{self.i1}@property''',
            (
                f'''{self.i1}def {model_type.plural()}(self) ->'''
                f''' Generator[{return_type}, None, None]:'''
            ),
            (
                f'''{self.i2}"""Yield {type_.singular().pascal_case()} '''
                f'''from {type_.plural().pascal_case()}."""'''
            ),
        ]

        if self.type_ != type_:
            _code.extend(
                [
                    (
                        f'''{self.i2}from {model_import_data.get('object_module')} '''
                        f'''import {model_import_data.get('object_collection_class')}'''
                    ),
                    '',
                ]
            )

        # Custom logic to ensure that when iterating over the associated indicators or associated
        # groups then the item currently being iterated over is not included in the results.
        if (
            self.type_ in ['indicators', 'groups', 'artifacts', 'cases']
            and model_type == f'associated_{self.type_}'
        ):
            _code.extend(
                [
                    f'''{self.i2}# Ensure the current item is not returned as a association''',
                    (
                        f'''{self.i2}for {type_.singular()} in self._iterate_over_sublist'''
                        f'''({model_import_data.get('object_collection_class')}):'''
                        '''  # type: ignore'''
                    ),
                ]
            )
            if self.type_ == 'indicator':
                _code.extend(
                    [
                        (
                            f'''{self.i3}if {type_.singular()}.model.summary == '''
                            '''self.model.summary:'''
                        ),
                    ]
                )
            else:
                _code.extend(
                    [
                        (f'''{self.i3}if {type_.singular()}.model.id == self.model.id:'''),
                    ]
                )
            _code.extend(
                [
                    f'''{self.i4}continue''',
                    f'''{self.i3}yield {type_.singular()}  # type: ignore''',
                ]
            )
        else:
            _code.extend(
                [
                    (
                        f'''{self.i2}yield from self._iterate_over_sublist'''
                        f'''({model_import_data.get('object_collection_class')})'''
                        '''  # type: ignore'''
                    ),
                ]
            )
        _code.extend(
            [
                '',
                '',
            ]
        )
        return '\n'.join(_code)

    def gen_container_class(self) -> str:
        """Generate the Container Model

        class Artifacts(ObjectCollectionABC):
            '''Artifacts Collection.

            # Example of params input
            {
                'result_limit': 100,  # Limit the retrieved results.
                'result_start': 10,  # Starting count used for pagination.
                'fields': ['caseId', 'summary']  # Select additional return fields.
            }

            Arg:
                session (Session): Session object configured with TC API Auth.
                tql_filters (list): List of TQL filters.
                params (dict): Additional query params (see example above).
            '''
        """
        return '\n'.join(
            [
                '',
                f'''class {self.type_.plural().pascal_case()}(ObjectCollectionABC):''',
                f'''{self.i1}"""{self.type_.plural().pascal_case()} Collection.''',
                '',
                f'''{self.i1}# Example of params input''',
                f'''{self.i1}{{''',
                f'''{self.i2}'result_limit': 100,  # Limit the retrieved results.''',
                f'''{self.i2}'result_start': 10,  # Starting count used for pagination.''',
                f'''{self.i2}'fields': ['caseId', 'summary']  # Select additional return fields.''',
                f'''{self.i1}}}''',
                '',
                f'''{self.i1}Args:''',
                f'''{self.i2}session (Session): Session object configured with TC API Auth.''',
                f'''{self.i2}tql_filters (list): List of TQL filters.''',
                f'''{self.i2}params (dict): Additional query params (see example above).''',
                f'{self.i1}"""',
                '',
                '',
            ]
        )

    def gen_container_methods(self) -> str:
        """Return the container methods.

        def __init__ ...

        def __iter__ ...

        def filter ...

        """
        _code = ''
        # generate __init__ method
        _code += self._gen_code_container_init_method()

        # generate __iter__ method
        _code += self._gen_code_container_iter_method()

        # generate __next__ method
        # _code += self._gen_code_container_next_method()

        # generate api_endpoint property method
        _code += self._gen_code_api_endpoint_property()

        # generate filter property method
        _code += self._gen_code_container_filter_property()

        if self.type_.lower() == 'indicators':
            _code += self._gen_code_deleted_method()

        return _code

    def gen_doc_string(self) -> str:
        """Generate doc string."""
        # return (
        #     f'"""{self.type_.singular().pascal_case()} '
        #     f'/ {self.type_.plural().pascal_case()} Object"""\n'
        # )
        return '"""TcEx Framework Module"""\n'

    def gen_object_class(self) -> str:
        """Generate the Object Model

        class Artifact(ObjectABC):
            '''Case Management Artifact

            Arg:
                case_id (int, kwargs): The **case id** for the Artifact.
                case_xid (str, kwargs): The **case xid** for the Artifact.
                derived_link (bool, kwargs): Flag to specify if this artifact should be used for
                    potentially associated cases or not.
                ...
            '''
        """
        args = GenerateArgs(self.type_).gen_args()
        return '\n'.join(
            [
                '',
                f'''class {self.type_.singular().pascal_case()}(ObjectABC):''',
                f'''{self.i1}"""{self.type_.plural().pascal_case()} Object.''',
                '',
                f'{args}',
                f'{self.i1}"""',
                '',
                '',
            ]
        )

    def gen_object_methods(self) -> str:
        """Return the container methods.

        def __init__ ...

        def __iter__ ...

        def filter ...

        """
        _code = ''

        # generate __init__ method
        _code += self._gen_code_object_init_method()

        # generate api_endpoint property method
        _code += self._gen_code_api_endpoint_property()

        # generate model property
        _code += self._gen_code_object_model_property()

        # skip object that don't require as_entity method
        if self.type_ not in [
            'case_attributes',
            'victim_attributes',
            'group_attributes',
            'indicator_attributes',
            'security_labels',
            'tags',
            'attribute_types',
            'owner_roles',
            'owners',
            'system_roles',
            'user_groups',
            'users',
        ]:
            # generate as_entity property method
            _code += self._gen_code_object_as_entity_property_method()

        # skip object that don't require remove method
        if self.type_ in [
            'groups',
            'indicators',
            'security_labels',
            'tags',
        ]:
            # generate remove property method
            _code += self._gen_code_object_remove_method()

        # generate group specific methods
        if self.type_.lower() == 'groups':
            _code += self._gen_code_group_methods()

        # get NON read-only properties of endpoint (OPTIONS: /v3/<object>)
        add_properties = [prop.name for prop in self._prop_models if prop.read_only is False]

        # generate artifacts property method
        if 'artifacts' in add_properties:
            _code += self._gen_code_object_type_property_method('artifacts')

        # generate victim assets property method
        if 'assets' in add_properties:
            _code += self._gen_code_object_type_property_method('victim_assets')

        # generate associated_case property method
        if 'associatedArtifacts' in add_properties:
            _code += self._gen_code_object_type_property_method('artifacts', 'associated_artifacts')

        # generate associated_case property method
        if 'associatedCases' in add_properties:
            _code += self._gen_code_object_type_property_method('cases', 'associated_cases')

        # generate associated_group property method
        if 'associatedGroups' in add_properties:
            _code += self._gen_code_object_type_property_method('groups', 'associated_groups')

        # generate associated_indicator property method
        if 'associatedIndicators' in add_properties:
            _code += self._gen_code_object_type_property_method(
                'indicators', 'associated_indicators'
            )

        # generate associated_victim_asset property method
        if 'associatedVictimAssets' in add_properties:
            _code += self._gen_code_object_type_property_method(
                'victim_assets', 'associated_victim_assets'
            )

        # generate attributes property method
        if 'attributes' in add_properties:
            _code += self._gen_code_object_type_property_method('attributes')

        # generate cases property method
        if 'cases' in add_properties:
            _code += self._gen_code_object_type_property_method('cases')

        # generate notes property method
        if 'notes' in add_properties:
            _code += self._gen_code_object_type_property_method('notes')

        # generate security_labels property method
        if 'securityLabels' in add_properties:
            _code += self._gen_code_object_type_property_method('security_labels')

        # generate tags property method
        if 'tags' in add_properties:
            _code += self._gen_code_object_type_property_method('tags')

        # generate tasks property method
        if 'tasks' in add_properties:
            _code += self._gen_code_object_type_property_method('tasks')

        #
        # Stage Method
        #

        # generate stage_artifact method
        if 'artifacts' in add_properties:
            _code += self._gen_code_object_stage_type_method('artifacts')

        # [custom] generate stage assignee method
        if self.type_.lower() in ['cases', 'tasks'] and 'assignee' in add_properties:
            _code += self._gen_code_object_stage_assignee()

        # generate stage_asset method
        if 'assets' in add_properties:
            _code += self._gen_code_object_stage_type_method('victim_assets')

        # generate stage_associated_group method
        if 'associatedCases' in add_properties:
            _code += self._gen_code_object_stage_type_method('cases', 'associated_cases')

        if 'associatedArtifacts' in add_properties:
            _code += self._gen_code_object_stage_type_method('artifacts', 'associated_artifacts')

        # victims have associatedGroups but groups must be
        # associated to the asset not the victim object.
        if 'associatedGroups' in add_properties and self.type_.lower() not in ['victims']:
            _code += self._gen_code_object_stage_type_method('groups', 'associated_groups')

        # generate stage_associated_group method
        if 'associatedVictimAssets' in add_properties:
            _code += self._gen_code_object_stage_type_method(
                'victim_assets', 'associated_victim_assets'
            )

        # generate stage_associated_indicator method
        if 'associatedIndicators' in add_properties and self.type_ != 'indicators':
            _code += self._gen_code_object_stage_type_method('indicators', 'associated_indicators')

        # generate stage_attribute method
        if 'attributes' in add_properties:
            _code += self._gen_code_object_stage_type_method('attributes')

        # generate stage_case method
        if 'cases' in add_properties:
            _code += self._gen_code_object_stage_type_method('cases')

        # generate stage_file_actions method
        if 'fileActions' in add_properties:
            _code += self._gen_code_object_stage_type_method('file_actions')

        # generate stage_file_occurrences method
        if 'fileOccurrences' in add_properties:
            _code += self._gen_code_object_stage_type_method('file_occurrences')

        # generate stage_note method
        if 'notes' in add_properties:
            _code += self._gen_code_object_stage_type_method('notes')

        # generate stage_security_labels method
        if 'securityLabels' in add_properties:
            _code += self._gen_code_object_stage_type_method('security_labels')

        # generate stage_tag method
        if 'tags' in add_properties:
            _code += self._gen_code_object_stage_type_method('tags')

        # generate stage_task method
        if 'tasks' in add_properties:
            _code += self._gen_code_object_stage_type_method('tasks')

        # generate stage_task method
        if 'userAccess' in add_properties:
            _code += self._gen_code_object_stage_type_method('users', 'user_access')

        return _code

    def update_requirements(
        self,
        type_: SnakeString,
        filename: str,
        classes: list[str],
        from_: str = 'first-party',
    ):
        """Return the requirements code."""
        class_string = ', '.join(classes)
        self.requirements[from_].append(  # type: ignore
            f'from {self.tap(type_)}.{type_.plural()}.{filename} import {class_string}'
        )



"""TcEx Framework Module"""
# standard library
from enum import Enum

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql import Tql
from tcex.api.tc.v3.tql.tql_operator import TqlOperator
from tcex.api.tc.v3.tql.tql_type import TqlType


class VictimFilter(FilterABC):
    """Filter Object for Victims"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.VICTIMS.value

    def asset_name(self, operator: Enum, asset_name: list | str):
        """Filter Asset Name based on **assetName** keyword.

        Args:
            operator: The operator enum for the filter.
            asset_name: The asset name assigned to a victim.
        """
        if isinstance(asset_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('assetName', operator, asset_name, TqlType.STRING)

    def asset_type(self, operator: Enum, asset_type: int | list):
        """Filter Asset Type ID based on **assetType** keyword.

        Args:
            operator: The operator enum for the filter.
            asset_type: The asset type ID assigned to a victim.
        """
        if isinstance(asset_type, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('assetType', operator, asset_type, TqlType.INTEGER)

    def asset_typename(self, operator: Enum, asset_typename: list | str):
        """Filter Asset Type Name based on **assetTypename** keyword.

        Args:
            operator: The operator enum for the filter.
            asset_typename: The asset type name assigned to a victim.
        """
        if isinstance(asset_typename, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('assetTypename', operator, asset_typename, TqlType.STRING)

    def attribute(self, operator: Enum, attribute: list | str):
        """Filter attribute based on **attribute** keyword.

        Args:
            operator: The operator enum for the filter.
            attribute: No description provided.
        """
        if isinstance(attribute, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('attribute', operator, attribute, TqlType.STRING)

    def description(self, operator: Enum, description: list | str):
        """Filter Description based on **description** keyword.

        Args:
            operator: The operator enum for the filter.
            description: The description of the victim.
        """
        if isinstance(description, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('description', operator, description, TqlType.STRING)

    @property
    def has_attribute(self):
        """Return **VictimAttributeFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.victim_attributes.victim_attribute_filter import VictimAttributeFilter

        attributes = VictimAttributeFilter(Tql())
        self._tql.add_filter('hasAttribute', TqlOperator.EQ, attributes, TqlType.SUB_QUERY)
        return attributes

    @property
    def has_group(self):
        """Return **GroupFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.groups.group_filter import GroupFilter

        groups = GroupFilter(Tql())
        self._tql.add_filter('hasGroup', TqlOperator.EQ, groups, TqlType.SUB_QUERY)
        return groups

    @property
    def has_indicator(self):
        """Return **IndicatorFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.indicators.indicator_filter import IndicatorFilter

        indicators = IndicatorFilter(Tql())
        self._tql.add_filter('hasIndicator', TqlOperator.EQ, indicators, TqlType.SUB_QUERY)
        return indicators

    @property
    def has_security_label(self):
        """Return **SecurityLabel** for further filtering."""
        # first-party
        from tcex.api.tc.v3.security_labels.security_label_filter import SecurityLabelFilter

        security_labels = SecurityLabelFilter(Tql())
        self._tql.add_filter('hasSecurityLabel', TqlOperator.EQ, security_labels, TqlType.SUB_QUERY)
        return security_labels

    @property
    def has_tag(self):
        """Return **TagFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.tags.tag_filter import TagFilter

        tags = TagFilter(Tql())
        self._tql.add_filter('hasTag', TqlOperator.EQ, tags, TqlType.SUB_QUERY)
        return tags

    @property
    def has_victim(self):
        """Return **VictimFilter** for further filtering."""
        victims = VictimFilter(Tql())
        self._tql.add_filter('hasVictim', TqlOperator.EQ, victims, TqlType.SUB_QUERY)
        return victims

    @property
    def has_victim_asset(self):
        """Return **VictimAssetFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.victim_assets.victim_asset_filter import VictimAssetFilter

        victim_assets = VictimAssetFilter(Tql())
        self._tql.add_filter('hasVictimAsset', TqlOperator.EQ, victim_assets, TqlType.SUB_QUERY)
        return victim_assets

    def id(self, operator: Enum, id: int | list):  # pylint: disable=redefined-builtin
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the victim.
        """
        if isinstance(id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def name(self, operator: Enum, name: list | str):
        """Filter Name based on **name** keyword.

        Args:
            operator: The operator enum for the filter.
            name: The name of the victim.
        """
        if isinstance(name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('name', operator, name, TqlType.STRING)

    def nationality(self, operator: Enum, nationality: list | str):
        """Filter Nationality based on **nationality** keyword.

        Args:
            operator: The operator enum for the filter.
            nationality: The nationality of the victim.
        """
        if isinstance(nationality, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('nationality', operator, nationality, TqlType.STRING)

    def organization(self, operator: Enum, organization: list | str):
        """Filter Organization based on **organization** keyword.

        Args:
            operator: The operator enum for the filter.
            organization: The organization of the victim.
        """
        if isinstance(organization, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('organization', operator, organization, TqlType.STRING)

    def owner(self, operator: Enum, owner: int | list):
        """Filter Owner ID based on **owner** keyword.

        Args:
            operator: The operator enum for the filter.
            owner: The owner ID of the victim.
        """
        if isinstance(owner, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('owner', operator, owner, TqlType.INTEGER)

    def owner_name(self, operator: Enum, owner_name: list | str):
        """Filter Owner Name based on **ownerName** keyword.

        Args:
            operator: The operator enum for the filter.
            owner_name: The owner name of the victim.
        """
        if isinstance(owner_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('ownerName', operator, owner_name, TqlType.STRING)

    def security_label(self, operator: Enum, security_label: list | str):
        """Filter Security Label based on **securityLabel** keyword.

        Args:
            operator: The operator enum for the filter.
            security_label: The name of a security label applied to the victim.
        """
        if isinstance(security_label, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('securityLabel', operator, security_label, TqlType.STRING)

    def sub_org(self, operator: Enum, sub_org: list | str):
        """Filter Sub-organization based on **subOrg** keyword.

        Args:
            operator: The operator enum for the filter.
            sub_org: The sub-organization of the victim.
        """
        if isinstance(sub_org, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('subOrg', operator, sub_org, TqlType.STRING)

    def summary(self, operator: Enum, summary: list | str):
        """Filter Summary based on **summary** keyword.

        Args:
            operator: The operator enum for the filter.
            summary: The name of the victim.
        """
        if isinstance(summary, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('summary', operator, summary, TqlType.STRING)

    def tag(self, operator: Enum, tag: list | str):
        """Filter Tag based on **tag** keyword.

        Args:
            operator: The operator enum for the filter.
            tag: The name of a tag applied to the victim.
        """
        if isinstance(tag, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('tag', operator, tag, TqlType.STRING)

    def tag_owner(self, operator: Enum, tag_owner: int | list):
        """Filter Tag Owner ID based on **tagOwner** keyword.

        Args:
            operator: The operator enum for the filter.
            tag_owner: The owner ID of a tag applied to the victim.
        """
        if isinstance(tag_owner, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('tagOwner', operator, tag_owner, TqlType.INTEGER)

    def tag_owner_name(self, operator: Enum, tag_owner_name: list | str):
        """Filter Tag Owner Name based on **tagOwnerName** keyword.

        Args:
            operator: The operator enum for the filter.
            tag_owner_name: The owner name of a tag applied to the victim.
        """
        if isinstance(tag_owner_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('tagOwnerName', operator, tag_owner_name, TqlType.STRING)

    def work_location(self, operator: Enum, work_location: list | str):
        """Filter Work Location based on **workLocation** keyword.

        Args:
            operator: The operator enum for the filter.
            work_location: The work location of the victim.
        """
        if isinstance(work_location, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('workLocation', operator, work_location, TqlType.STRING)



"""TcEx Framework Module"""
# standard library
from datetime import datetime
from enum import Enum

# third-party
from arrow import Arrow

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql import Tql
from tcex.api.tc.v3.tql.tql_operator import TqlOperator
from tcex.api.tc.v3.tql.tql_type import TqlType


class GroupFilter(FilterABC):
    """Filter Object for Groups"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.GROUPS.value

    def associated_indicator(self, operator: Enum, associated_indicator: int | list):
        """Filter associatedIndicator based on **associatedIndicator** keyword.

        Args:
            operator: The operator enum for the filter.
            associated_indicator: No description provided.
        """
        if isinstance(associated_indicator, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('associatedIndicator', operator, associated_indicator, TqlType.INTEGER)

    def attribute(self, operator: Enum, attribute: list | str):
        """Filter attribute based on **attribute** keyword.

        Args:
            operator: The operator enum for the filter.
            attribute: No description provided.
        """
        if isinstance(attribute, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('attribute', operator, attribute, TqlType.STRING)

    def child_group(self, operator: Enum, child_group: int | list):
        """Filter childGroup based on **childGroup** keyword.

        Args:
            operator: The operator enum for the filter.
            child_group: No description provided.
        """
        if isinstance(child_group, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('childGroup', operator, child_group, TqlType.INTEGER)

    def created_by(self, operator: Enum, created_by: list | str):
        """Filter Created By based on **createdBy** keyword.

        Args:
            operator: The operator enum for the filter.
            created_by: The user who created the group.
        """
        if isinstance(created_by, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('createdBy', operator, created_by, TqlType.STRING)

    def date_added(self, operator: Enum, date_added: Arrow | datetime | int | str):
        """Filter Date Added based on **dateAdded** keyword.

        Args:
            operator: The operator enum for the filter.
            date_added: The date the group was added to the system.
        """
        date_added = self.util.any_to_datetime(date_added).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('dateAdded', operator, date_added, TqlType.STRING)

    def document_date_added(
        self, operator: Enum, document_date_added: Arrow | datetime | int | str
    ):
        """Filter Date Added (Document) based on **documentDateAdded** keyword.

        Args:
            operator: The operator enum for the filter.
            document_date_added: The date the document was added.
        """
        document_date_added = self.util.any_to_datetime(document_date_added).strftime(
            '%Y-%m-%d %H:%M:%S'
        )
        self._tql.add_filter('documentDateAdded', operator, document_date_added, TqlType.STRING)

    def document_filename(self, operator: Enum, document_filename: list | str):
        """Filter Filename (Document) based on **documentFilename** keyword.

        Args:
            operator: The operator enum for the filter.
            document_filename: The file name of the document.
        """
        if isinstance(document_filename, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('documentFilename', operator, document_filename, TqlType.STRING)

    def document_filesize(self, operator: Enum, document_filesize: int | list):
        """Filter File Size (Document) based on **documentFilesize** keyword.

        Args:
            operator: The operator enum for the filter.
            document_filesize: The filesize of the document.
        """
        if isinstance(document_filesize, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('documentFilesize', operator, document_filesize, TqlType.INTEGER)

    def document_status(self, operator: Enum, document_status: list | str):
        """Filter Status (Document) based on **documentStatus** keyword.

        Args:
            operator: The operator enum for the filter.
            document_status: The status of the document.
        """
        if isinstance(document_status, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('documentStatus', operator, document_status, TqlType.STRING)

    def document_type(self, operator: Enum, document_type: list | str):
        """Filter Type (Document) based on **documentType** keyword.

        Args:
            operator: The operator enum for the filter.
            document_type: The type of document.
        """
        if isinstance(document_type, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('documentType', operator, document_type, TqlType.STRING)

    def downvote_count(self, operator: Enum, downvote_count: int | list):
        """Filter Downvote Count based on **downvoteCount** keyword.

        Args:
            operator: The operator enum for the filter.
            downvote_count: The number of downvotes the group has received.
        """
        if isinstance(downvote_count, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('downvoteCount', operator, downvote_count, TqlType.INTEGER)

    def email_date(self, operator: Enum, email_date: Arrow | datetime | int | str):
        """Filter Date (Email) based on **emailDate** keyword.

        Args:
            operator: The operator enum for the filter.
            email_date: The date of the email.
        """
        email_date = self.util.any_to_datetime(email_date).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('emailDate', operator, email_date, TqlType.STRING)

    def email_from(self, operator: Enum, email_from: list | str):
        """Filter From (Email) based on **emailFrom** keyword.

        Args:
            operator: The operator enum for the filter.
            email_from: The 'from' field of the email.
        """
        if isinstance(email_from, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('emailFrom', operator, email_from, TqlType.STRING)

    def email_score(self, operator: Enum, email_score: int | list):
        """Filter Score (Email) based on **emailScore** keyword.

        Args:
            operator: The operator enum for the filter.
            email_score: The score of the email.
        """
        if isinstance(email_score, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('emailScore', operator, email_score, TqlType.INTEGER)

    def email_score_includes_body(self, operator: Enum, email_score_includes_body: bool):
        """Filter Score Includes Body (Email) based on **emailScoreIncludesBody** keyword.

        Args:
            operator: The operator enum for the filter.
            email_score_includes_body: A true/false indicating if the body was included in the
                scoring of the email.
        """
        self._tql.add_filter(
            'emailScoreIncludesBody', operator, email_score_includes_body, TqlType.BOOLEAN
        )

    def email_subject(self, operator: Enum, email_subject: list | str):
        """Filter Subject (Email) based on **emailSubject** keyword.

        Args:
            operator: The operator enum for the filter.
            email_subject: The subject of the email.
        """
        if isinstance(email_subject, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('emailSubject', operator, email_subject, TqlType.STRING)

    def event_date(self, operator: Enum, event_date: Arrow | datetime | int | str):
        """Filter Event Date based on **eventDate** keyword.

        Args:
            operator: The operator enum for the filter.
            event_date: The event date of the group.
        """
        event_date = self.util.any_to_datetime(event_date).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('eventDate', operator, event_date, TqlType.STRING)

    def generated_report(self, operator: Enum, generated_report: bool):
        """Filter Generated (Report) based on **generatedReport** keyword.

        Args:
            operator: The operator enum for the filter.
            generated_report: Boolean flag indicating if the Report was auto-generated.
        """
        self._tql.add_filter('generatedReport', operator, generated_report, TqlType.BOOLEAN)

    @property
    def has_artifact(self):
        """Return **ArtifactFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.artifacts.artifact_filter import ArtifactFilter

        artifacts = ArtifactFilter(Tql())
        self._tql.add_filter('hasArtifact', TqlOperator.EQ, artifacts, TqlType.SUB_QUERY)
        return artifacts

    @property
    def has_attribute(self):
        """Return **GroupAttributeFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.group_attributes.group_attribute_filter import GroupAttributeFilter

        attributes = GroupAttributeFilter(Tql())
        self._tql.add_filter('hasAttribute', TqlOperator.EQ, attributes, TqlType.SUB_QUERY)
        return attributes

    @property
    def has_case(self):
        """Return **CaseFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.cases.case_filter import CaseFilter

        cases = CaseFilter(Tql())
        self._tql.add_filter('hasCase', TqlOperator.EQ, cases, TqlType.SUB_QUERY)
        return cases

    @property
    def has_group(self):
        """Return **GroupFilter** for further filtering."""
        groups = GroupFilter(Tql())
        self._tql.add_filter('hasGroup', TqlOperator.EQ, groups, TqlType.SUB_QUERY)
        return groups

    @property
    def has_indicator(self):
        """Return **IndicatorFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.indicators.indicator_filter import IndicatorFilter

        indicators = IndicatorFilter(Tql())
        self._tql.add_filter('hasIndicator', TqlOperator.EQ, indicators, TqlType.SUB_QUERY)
        return indicators

    def has_intel_query(self, operator: Enum, has_intel_query: int | list):
        """Filter Associated User Queries based on **hasIntelQuery** keyword.

        Args:
            operator: The operator enum for the filter.
            has_intel_query: A nested query for association to User Queries.
        """
        if isinstance(has_intel_query, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('hasIntelQuery', operator, has_intel_query, TqlType.INTEGER)

    @property
    def has_security_label(self):
        """Return **SecurityLabel** for further filtering."""
        # first-party
        from tcex.api.tc.v3.security_labels.security_label_filter import SecurityLabelFilter

        security_labels = SecurityLabelFilter(Tql())
        self._tql.add_filter('hasSecurityLabel', TqlOperator.EQ, security_labels, TqlType.SUB_QUERY)
        return security_labels

    @property
    def has_tag(self):
        """Return **TagFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.tags.tag_filter import TagFilter

        tags = TagFilter(Tql())
        self._tql.add_filter('hasTag', TqlOperator.EQ, tags, TqlType.SUB_QUERY)
        return tags

    @property
    def has_victim(self):
        """Return **VictimFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.victims.victim_filter import VictimFilter

        victims = VictimFilter(Tql())
        self._tql.add_filter('hasVictim', TqlOperator.EQ, victims, TqlType.SUB_QUERY)
        return victims

    @property
    def has_victim_asset(self):
        """Return **VictimAssetFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.victim_assets.victim_asset_filter import VictimAssetFilter

        victim_assets = VictimAssetFilter(Tql())
        self._tql.add_filter('hasVictimAsset', TqlOperator.EQ, victim_assets, TqlType.SUB_QUERY)
        return victim_assets

    def id(self, operator: Enum, id: int | list):  # pylint: disable=redefined-builtin
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the group.
        """
        if isinstance(id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def is_group(self, operator: Enum, is_group: bool):
        """Filter isGroup based on **isGroup** keyword.

        Args:
            operator: The operator enum for the filter.
            is_group: No description provided.
        """
        self._tql.add_filter('isGroup', operator, is_group, TqlType.BOOLEAN)

    def last_modified(self, operator: Enum, last_modified: Arrow | datetime | int | str):
        """Filter Last Modified based on **lastModified** keyword.

        Args:
            operator: The operator enum for the filter.
            last_modified: The date the group was last modified.
        """
        last_modified = self.util.any_to_datetime(last_modified).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('lastModified', operator, last_modified, TqlType.STRING)

    def owner(self, operator: Enum, owner: int | list):
        """Filter Owner ID based on **owner** keyword.

        Args:
            operator: The operator enum for the filter.
            owner: The Owner ID for the group.
        """
        if isinstance(owner, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('owner', operator, owner, TqlType.INTEGER)

    def owner_name(self, operator: Enum, owner_name: list | str):
        """Filter Owner Name based on **ownerName** keyword.

        Args:
            operator: The operator enum for the filter.
            owner_name: The owner name for the group.
        """
        if isinstance(owner_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('ownerName', operator, owner_name, TqlType.STRING)

    def parent_group(self, operator: Enum, parent_group: int | list):
        """Filter parentGroup based on **parentGroup** keyword.

        Args:
            operator: The operator enum for the filter.
            parent_group: No description provided.
        """
        if isinstance(parent_group, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('parentGroup', operator, parent_group, TqlType.INTEGER)

    def security_label(self, operator: Enum, security_label: list | str):
        """Filter Security Label based on **securityLabel** keyword.

        Args:
            operator: The operator enum for the filter.
            security_label: The name of a security label applied to the group.
        """
        if isinstance(security_label, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('securityLabel', operator, security_label, TqlType.STRING)

    def signature_date_added(
        self, operator: Enum, signature_date_added: Arrow | datetime | int | str
    ):
        """Filter Date Added (Signature) based on **signatureDateAdded** keyword.

        Args:
            operator: The operator enum for the filter.
            signature_date_added: The date the signature was added.
        """
        signature_date_added = self.util.any_to_datetime(signature_date_added).strftime(
            '%Y-%m-%d %H:%M:%S'
        )
        self._tql.add_filter('signatureDateAdded', operator, signature_date_added, TqlType.STRING)

    def signature_filename(self, operator: Enum, signature_filename: list | str):
        """Filter Filename (Signature) based on **signatureFilename** keyword.

        Args:
            operator: The operator enum for the filter.
            signature_filename: The file name of the signature.
        """
        if isinstance(signature_filename, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('signatureFilename', operator, signature_filename, TqlType.STRING)

    def signature_type(self, operator: Enum, signature_type: list | str):
        """Filter Type (Signature) based on **signatureType** keyword.

        Args:
            operator: The operator enum for the filter.
            signature_type: The type of signature.
        """
        if isinstance(signature_type, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('signatureType', operator, signature_type, TqlType.STRING)

    def status(self, operator: Enum, status: list | str):
        """Filter Status based on **status** keyword.

        Args:
            operator: The operator enum for the filter.
            status: Status of the group.
        """
        if isinstance(status, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('status', operator, status, TqlType.STRING)

    def summary(self, operator: Enum, summary: list | str):
        """Filter Summary based on **summary** keyword.

        Args:
            operator: The operator enum for the filter.
            summary: The summary (name) of the group.
        """
        if isinstance(summary, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('summary', operator, summary, TqlType.STRING)

    def tag(self, operator: Enum, tag: list | str):
        """Filter Tag based on **tag** keyword.

        Args:
            operator: The operator enum for the filter.
            tag: The name of a tag applied to the group.
        """
        if isinstance(tag, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('tag', operator, tag, TqlType.STRING)

    def tag_owner(self, operator: Enum, tag_owner: int | list):
        """Filter Tag Owner ID based on **tagOwner** keyword.

        Args:
            operator: The operator enum for the filter.
            tag_owner: The ID of the owner of a tag.
        """
        if isinstance(tag_owner, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('tagOwner', operator, tag_owner, TqlType.INTEGER)

    def tag_owner_name(self, operator: Enum, tag_owner_name: list | str):
        """Filter Tag Owner Name based on **tagOwnerName** keyword.

        Args:
            operator: The operator enum for the filter.
            tag_owner_name: The name of the owner of a tag.
        """
        if isinstance(tag_owner_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('tagOwnerName', operator, tag_owner_name, TqlType.STRING)

    def task_assignee(self, operator: Enum, task_assignee: list | str):
        """Filter Assignee (Task) based on **taskAssignee** keyword.

        Args:
            operator: The operator enum for the filter.
            task_assignee: The assignee of the task.
        """
        if isinstance(task_assignee, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('taskAssignee', operator, task_assignee, TqlType.STRING)

    def task_assignee_pseudo(self, operator: Enum, task_assignee_pseudo: list | str):
        """Filter Assignee Pseudonym (Task) based on **taskAssigneePseudo** keyword.

        Args:
            operator: The operator enum for the filter.
            task_assignee_pseudo: The pseudonym of the assignee of the task.
        """
        if isinstance(task_assignee_pseudo, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('taskAssigneePseudo', operator, task_assignee_pseudo, TqlType.STRING)

    def task_date_added(self, operator: Enum, task_date_added: Arrow | datetime | int | str):
        """Filter Date Added (Task) based on **taskDateAdded** keyword.

        Args:
            operator: The operator enum for the filter.
            task_date_added: The date the task was added.
        """
        task_date_added = self.util.any_to_datetime(task_date_added).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('taskDateAdded', operator, task_date_added, TqlType.STRING)

    def task_due_date(self, operator: Enum, task_due_date: Arrow | datetime | int | str):
        """Filter Due Date (Task) based on **taskDueDate** keyword.

        Args:
            operator: The operator enum for the filter.
            task_due_date: The due date of a task.
        """
        task_due_date = self.util.any_to_datetime(task_due_date).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('taskDueDate', operator, task_due_date, TqlType.STRING)

    def task_escalated(self, operator: Enum, task_escalated: bool):
        """Filter Escalated (Task) based on **taskEscalated** keyword.

        Args:
            operator: The operator enum for the filter.
            task_escalated: A flag indicating if a task has been escalated.
        """
        self._tql.add_filter('taskEscalated', operator, task_escalated, TqlType.BOOLEAN)

    def task_escalation_date(
        self, operator: Enum, task_escalation_date: Arrow | datetime | int | str
    ):
        """Filter Escalation Date (Task) based on **taskEscalationDate** keyword.

        Args:
            operator: The operator enum for the filter.
            task_escalation_date: The escalation date of a task.
        """
        task_escalation_date = self.util.any_to_datetime(task_escalation_date).strftime(
            '%Y-%m-%d %H:%M:%S'
        )
        self._tql.add_filter('taskEscalationDate', operator, task_escalation_date, TqlType.STRING)

    def task_last_modified(self, operator: Enum, task_last_modified: Arrow | datetime | int | str):
        """Filter Last Modified based on **taskLastModified** keyword.

        Args:
            operator: The operator enum for the filter.
            task_last_modified: The date the group was last modified.
        """
        task_last_modified = self.util.any_to_datetime(task_last_modified).strftime(
            '%Y-%m-%d %H:%M:%S'
        )
        self._tql.add_filter('taskLastModified', operator, task_last_modified, TqlType.STRING)

    def task_overdue(self, operator: Enum, task_overdue: bool):
        """Filter Overdue (Task) based on **taskOverdue** keyword.

        Args:
            operator: The operator enum for the filter.
            task_overdue: A flag indicating if a task has become overdue.
        """
        self._tql.add_filter('taskOverdue', operator, task_overdue, TqlType.BOOLEAN)

    def task_reminded(self, operator: Enum, task_reminded: bool):
        """Filter Reminded (Task) based on **taskReminded** keyword.

        Args:
            operator: The operator enum for the filter.
            task_reminded: A flag indicating if a task has been reminded.
        """
        self._tql.add_filter('taskReminded', operator, task_reminded, TqlType.BOOLEAN)

    def task_reminder_date(self, operator: Enum, task_reminder_date: Arrow | datetime | int | str):
        """Filter Reminder Date (Task) based on **taskReminderDate** keyword.

        Args:
            operator: The operator enum for the filter.
            task_reminder_date: The reminder date of a task.
        """
        task_reminder_date = self.util.any_to_datetime(task_reminder_date).strftime(
            '%Y-%m-%d %H:%M:%S'
        )
        self._tql.add_filter('taskReminderDate', operator, task_reminder_date, TqlType.STRING)

    def task_status(self, operator: Enum, task_status: list | str):
        """Filter Status (Task) based on **taskStatus** keyword.

        Args:
            operator: The operator enum for the filter.
            task_status: The status of the task.
        """
        if isinstance(task_status, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('taskStatus', operator, task_status, TqlType.STRING)

    def type(self, operator: Enum, type: int | list):  # pylint: disable=redefined-builtin
        """Filter Type based on **type** keyword.

        Args:
            operator: The operator enum for the filter.
            type: The ID of the group type.
        """
        if isinstance(type, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('type', operator, type, TqlType.INTEGER)

    def type_name(self, operator: Enum, type_name: list | str):
        """Filter Type Name based on **typeName** keyword.

        Args:
            operator: The operator enum for the filter.
            type_name: The name of the group type.
        """
        if isinstance(type_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('typeName', operator, type_name, TqlType.STRING)

    def upvote_count(self, operator: Enum, upvote_count: int | list):
        """Filter Upvote Count based on **upvoteCount** keyword.

        Args:
            operator: The operator enum for the filter.
            upvote_count: The number of upvotes the group has received.
        """
        if isinstance(upvote_count, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('upvoteCount', operator, upvote_count, TqlType.INTEGER)

    def victim_asset(self, operator: Enum, victim_asset: list | str):
        """Filter victimAsset based on **victimAsset** keyword.

        Args:
            operator: The operator enum for the filter.
            victim_asset: No description provided.
        """
        if isinstance(victim_asset, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('victimAsset', operator, victim_asset, TqlType.STRING)
